"""Test For loop implementation in IR generator."""

import pytest
from pecos.slr import CReg, For, Main, QReg
from pecos.slr.gen_codes.guppy.ir_generator import IRGuppyGenerator
from pecos.slr.misc import Comment
from pecos.slr.qeclib import qubit
from pecos.slr.qeclib.qubit.measures import Measure


def test_for_loop_range_basic() -> None:
    """Test basic for loop with range."""
    prog = Main(
        q := QReg("q", 5),
        # Apply H gate to each qubit using For loop
        For("i", 0, 5).Do(
            Comment("Apply H to qubit i"),
            # In real implementation, we'd use: qubit.H(q[i])
            # For now, just apply to q[0] as placeholder
            qubit.H(q[0]),
        ),
        Measure(q) > CReg("results", 5),
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check that for loop is generated
    assert "for i in range(0, 5):" in code
    assert "# Apply H to qubit i" in code


def test_for_loop_range_with_step() -> None:
    """Test for loop with custom step."""
    prog = Main(
        q := QReg("q", 10),
        # Apply X to every other qubit
        For("i", 0, 10, 2).Do(
            Comment("Apply X to even-indexed qubits"),
            qubit.X(q[0]),  # Placeholder
        ),
        Measure(q) > CReg("results", 10),
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check step parameter
    assert "for i in range(0, 10, 2):" in code


def test_for_loop_iterable() -> None:
    """Test for loop over an iterable."""
    prog = Main(
        q := QReg("q", 3),
        _indices := CReg("indices", 3),
        # For loop over a collection (conceptual)
        For("idx", "indices").Do(
            Comment("Process index from collection"),
            qubit.Y(q[0]),
        ),
        Measure(q) > CReg("results", 3),
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check iterable pattern
    assert "for idx in indices:" in code


def test_nested_for_loops() -> None:
    """Test nested for loops."""
    prog = Main(
        q := QReg("q", 9),  # 3x3 grid
        # Nested loops for 2D pattern
        For("i", 0, 3).Do(
            Comment("Outer loop"),
            For("j", 0, 3).Do(
                Comment("Inner loop"),
                Comment("Would apply operation to q[i*3 + j]"),
                qubit.H(q[0]),  # Placeholder
            ),
        ),
        Measure(q) > CReg("results", 9),
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check nested structure
    assert "for i in range(0, 3):" in code
    assert "for j in range(0, 3):" in code
    assert "# Outer loop" in code
    assert "# Inner loop" in code


def test_for_loop_with_quantum_operations() -> None:
    """Test for loop with quantum operations inside."""
    prog = Main(
        data := QReg("data", 4),
        ancilla := QReg("ancilla", 1),
        # Initialize data qubits
        For("i", 0, 4).Do(
            Comment("Initialize data qubit"),
            qubit.H(data[0]),  # Would be data[i]
        ),
        # Entangle with ancilla
        For("i", 0, 4).Do(
            Comment("Entangle with ancilla"),
            qubit.CX(data[0], ancilla[0]),  # Would be data[i]
        ),
        Measure(data) > CReg("data_results", 4),
        Measure(ancilla) > CReg("ancilla_result", 1),
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check multiple for loops
    assert code.count("for i in range(0, 4):") >= 2
    assert "quantum.h" in code
    assert "quantum.cx" in code


def test_for_loop_limitations() -> None:
    """Test current limitations of For loop implementation."""
    # The main limitation is that we can't use the loop variable
    # to index into quantum registers yet

    prog = Main(
        q := QReg("q", 5),
        For("i", 0, 5).Do(
            Comment("TODO: Need to support q[i] indexing"),
            Comment("Currently would need to unpack array first"),
        ),
        Measure(q) > CReg("c", 5),
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Document the limitation
    assert "for i in range(0, 5):" in code
    assert "TODO" in code


def test_for_error_in_qasm() -> None:
    """Test that For loops raise error in QASM generator."""
    from pecos.slr.gen_codes.gen_qasm import QASMGenerator

    prog = Main(
        q := QReg("q", 3),
        For("i", 0, 3).Do(
            qubit.H(q[0]),
        ),
        Measure(q) > CReg("c", 3),
    )

    gen = QASMGenerator(_internal=True)

    # Should raise NotImplementedError
    with pytest.raises(NotImplementedError) as exc_info:
        gen.generate_block(prog)

    assert "For loops are not supported in QASM" in str(exc_info.value)



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

"""QEC result analysis and post-processing utilities.

This module provides utilities for analyzing quantum error correction
results, including logical operator extraction, fidelity calculation,
and syndrome processing.
"""

from __future__ import annotations

import math
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence


def logical_x_from_data(d: int, data: Sequence[int]) -> int:
    """Extract logical X value from measurement data.

    For a surface code, the logical X operator is supported on the
    left column of data qubits (indices 0, d, 2d, ...).

    Args:
        d: Code distance.
        data: Measurement outcomes for d^2 data qubits.

    Returns:
        Logical X value (0 or 1) computed as XOR of left column.
    """
    if len(data) != d * d:
        msg = f"Expected {d*d} data qubits, got {len(data)}"
        raise ValueError(msg)

    result = 0
    for i in range(d):
        result ^= data[i * d]
    return result


def logical_z_from_data(d: int, data: Sequence[int]) -> int:
    """Extract logical Z value from measurement data.

    For a surface code, the logical Z operator is supported on the
    top row of data qubits (indices 0, 1, 2, ..., d-1).

    Args:
        d: Code distance.
        data: Measurement outcomes for d^2 data qubits.

    Returns:
        Logical Z value (0 or 1) computed as XOR of top row.
    """
    if len(data) != d * d:
        msg = f"Expected {d*d} data qubits, got {len(data)}"
        raise ValueError(msg)

    result = 0
    for i in range(d):
        result ^= data[i]
    return result


def logical_from_data(d: int, data: Sequence[int]) -> tuple[int, int]:
    """Extract both logical X and Z values from measurement data.

    Args:
        d: Code distance.
        data: Measurement outcomes for d^2 data qubits.

    Returns:
        Tuple of (logical_x, logical_z) values.
    """
    return logical_x_from_data(d, data), logical_z_from_data(d, data)


def logical_fidelity(
    outcomes: Sequence[Sequence[int]],
    d: int,
    basis: int,
    expected: int = 0,
) -> tuple[float, float]:
    """Compute logical fidelity from multiple measurement outcomes.

    Calculates the fraction of shots where the logical measurement
    matches the expected value, with binomial error bars.

    Args:
        outcomes: List of measurement outcomes, each with d^2 data qubits.
        d: Code distance.
        basis: Measurement basis (0=X, 1=Z).
        expected: Expected logical value (default 0).

    Returns:
        Tuple of (fidelity, error) where error is the standard deviation
        assuming binomial statistics: sqrt(f * (1-f) / num_shots).
    """
    if not outcomes:
        msg = "No outcomes provided"
        raise ValueError(msg)

    num_shots = len(outcomes)
    successes = 0

    for data in outcomes:
        logical = logical_x_from_data(d, data) if basis == 0 else logical_z_from_data(d, data)

        if logical == expected:
            successes += 1

    fidelity = successes / num_shots

    # Binomial error bar
    error = 0.0 if fidelity == 0.0 or fidelity == 1.0 else math.sqrt(fidelity * (1 - fidelity) / num_shots)

    return fidelity, error


def syndrome_difference(
    syndromes: Sequence[Sequence[int]],
) -> list[list[int]]:
    """Compute syndrome differences between consecutive rounds.

    For MWPM decoding, we need the syndrome changes (differences)
    between rounds rather than the raw syndromes. The first round
    is compared against all-zeros (initialization).

    Args:
        syndromes: List of syndrome measurements, one per round.
            Each syndrome is a sequence of stabilizer measurement results.

    Returns:
        List of syndrome differences. The i-th difference is
        syndromes[i] XOR syndromes[i-1] (with syndromes[-1] = 0).
    """
    if not syndromes:
        return []

    num_stab = len(syndromes[0])
    result = []

    # First round: difference from all-zeros
    result.append(list(syndromes[0]))

    # Subsequent rounds: difference from previous
    for i in range(1, len(syndromes)):
        diff = [syndromes[i][j] ^ syndromes[i - 1][j] for j in range(num_stab)]
        result.append(diff)

    return result


def syndrome_to_detection_events(
    syndromes: Sequence[Sequence[int]],
) -> list[tuple[int, int]]:
    """Convert syndrome history to detection event coordinates.

    Detection events are (stabilizer_index, round) pairs where
    the syndrome changed (difference is 1).

    Args:
        syndromes: List of syndrome measurements, one per round.

    Returns:
        List of (stabilizer_index, round) tuples for each detection event.
    """
    differences = syndrome_difference(syndromes)
    events = []

    for t, diff in enumerate(differences):
        for s, val in enumerate(diff):
            if val == 1:
                events.append((s, t))

    return events


def logical_error_rate(
    outcomes: Sequence[Sequence[int]],
    d: int,
    basis: int,
    expected: int = 0,
) -> tuple[float, float]:
    """Compute logical error rate from measurement outcomes.

    This is simply 1 - fidelity, with propagated error bars.

    Args:
        outcomes: List of measurement outcomes, each with d^2 data qubits.
        d: Code distance.
        basis: Measurement basis (0=X, 1=Z).
        expected: Expected logical value (default 0).

    Returns:
        Tuple of (error_rate, error_bar).
    """
    fidelity, error = logical_fidelity(outcomes, d, basis, expected)
    return 1 - fidelity, error


def lower_bound_fidelity(f1: float, f2: float) -> float:
    """Compute lower bound on true fidelity from two measurements.

    Uses the formula: bound = (4/5) * (f1 + f2) - (3/5)

    This provides a statistical lower bound accounting for
    measurement correlations when measuring in two different bases.

    Args:
        f1: Fidelity from first measurement basis.
        f2: Fidelity from second measurement basis.

    Returns:
        Lower bound on true fidelity.
    """
    return (4 / 5) * (f1 + f2) - (3 / 5)



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Integration tests for state vector quantum simulators using pure PECOS (no NumPy)."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

    from pecos.simulators.sim_class_types import StateVector

import json
from pathlib import Path

import pecos as pc
import pytest
from pecos.circuits import QuantumCircuit
from pecos.engines.hybrid_engine import HybridEngine
from pecos.error_models.generic_error_model import GenericErrorModel
from pecos.simulators import (
    MPS,
    CuStateVec,
    QuestStateVec,
    Qulacs,
    StateVec,
)
from pecos.testing import assert_allclose

str_to_sim = {
    "StateVec": StateVec,
    "Qulacs": Qulacs,
    "CuStateVec": CuStateVec,
    "MPS": MPS,
    "QuestStateVec": QuestStateVec,
}


def check_dependencies(
    simulator: str,
    **kwargs: object,
) -> Callable[[int], StateVector]:
    """Check if dependencies for a simulator are available and skip test if not.

    Args:
        simulator: Name of the simulator to check.
        **kwargs: Optional parameters to pass to the simulator constructor.

    Returns:
        A function that creates a simulator instance with the given parameters.
    """
    if simulator not in str_to_sim or str_to_sim[simulator] is None:
        pytest.skip(f"Requirements to test {simulator} are not met.")
    sim_class = str_to_sim[simulator]

    # Return a lambda that passes kwargs to the simulator constructor
    if kwargs:
        return lambda num_qubits: sim_class(num_qubits, **kwargs)
    return sim_class


def verify(simulator: str, qc: QuantumCircuit, final_vector: pc.Array) -> None:
    """Verify quantum circuit simulation results against expected state vector."""
    sim = check_dependencies(simulator)(len(qc.qudits))
    sim.run_circuit(qc)

    # Normalize vectors
    sim_vector_normalized = sim.vector / (pc.linalg.norm(sim.vector) or 1)
    final_vector_normalized = final_vector / (pc.linalg.norm(final_vector) or 1)

    phase = final_vector_normalized[0] / sim_vector_normalized[0] if pc.abs(sim_vector_normalized[0]) > 1e-10 else 1

    sim_vector_adjusted = sim_vector_normalized * phase

    # Use looser tolerance for simulators that use gate decompositions
    # QuestStateVec uses decompositions for RXX, RYY, RZZ which accumulate errors
    rtol = 1e-3 if simulator == "QuestStateVec" else 1e-5

    # Add absolute tolerance to handle near-zero values with numerical noise
    # MPS uses tensor network approximations that can introduce ~1e-15 errors
    # This prevents "inf" relative errors when comparing to exact 0
    atol = 1e-12

    assert_allclose(
        sim_vector_adjusted,
        final_vector_normalized,
        rtol=rtol,
        atol=atol,
        err_msg="State vectors do not match.",
    )


def check_measurement(
    simulator: str,
    qc: QuantumCircuit,
    final_results: dict[int, int] | None = None,
) -> None:
    """Check measurement results from quantum circuit simulation."""
    sim = check_dependencies(simulator)(len(qc.qudits))

    results = sim.run_circuit(qc)

    if final_results is not None:
        assert results == final_results

    state = 0
    for q, value in results.items():
        state += value * 2 ** (sim.num_qubits - 1 - q)
    final_vector = pc.zeros(shape=(2**sim.num_qubits,), dtype=pc.dtypes.complex128)
    final_vector[state] = 1

    abs_values_vector = [pc.abs(x) for x in sim.vector]

    assert pc.allclose(abs_values_vector, final_vector)


def compare_against_statevec(
    simulator: str,
    qc: QuantumCircuit,
    **sim_kwargs: object,
) -> None:
    """Compare simulator results against StateVec reference implementation.

    Args:
        simulator: Name of the simulator to test.
        qc: Quantum circuit to simulate.
        **sim_kwargs: Optional parameters passed to the simulator constructor.
            For MPS, use chi=32 or truncation_fidelity=0.999 for faster tests
            (cannot use both simultaneously).
    """
    statevec = StateVec(len(qc.qudits))
    statevec.run_circuit(qc)

    sim = check_dependencies(simulator, **sim_kwargs)(len(qc.qudits))
    sim.run_circuit(qc)

    # Use updated verify function
    verify(simulator, qc, statevec.vector)


def generate_random_state(seed: int | None = None) -> QuantumCircuit:
    """Generate a quantum circuit with random gates for testing."""
    pc.random.seed(seed)

    qc = QuantumCircuit()
    qc.append({"Init": {0, 1, 2, 3}})

    for _ in range(3):
        qc.append({"RZ": {0}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RZ": {1}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RZ": {2}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RZ": {3}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RXX": {(0, 1)}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RXX": {(0, 2)}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RXX": {(0, 3)}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RXX": {(1, 2)}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RXX": {(1, 3)}}, angles=(pc.f64.pi * pc.random.random(1)[0],))
        qc.append({"RXX": {(2, 3)}}, angles=(pc.f64.pi * pc.random.random(1)[0],))

    return qc


@pytest.mark.parametrize(
    "simulator",
    [
        "StateVec",
        "Qulacs",
        "CuStateVec",
        "MPS",
        "QuestStateVec",
    ],
)
def test_init(simulator: str) -> None:
    """Test quantum state initialization."""
    qc = QuantumCircuit()
    qc.append({"Init": {0, 1, 2, 3}})

    final_vector = pc.zeros(shape=(2**4,), dtype=pc.dtypes.complex128)
    final_vector[0] = 1

    verify(simulator, qc, final_vector)


@pytest.mark.parametrize(
    "simulator",
    [
        "StateVec",
        "Qulacs",
        "CuStateVec",
        "MPS",
        "QuestStateVec",
    ],
)
def test_H_measure(simulator: str) -> None:
    """Test Hadamard gate followed by measurement."""
    qc = QuantumCircuit()
    qc.append({"H": {0, 1, 2, 3}})
    qc.append({"Measure": {0, 1, 2, 3}})

    check_measurement(simulator, qc)


@pytest.mark.parametrize(
    "simulator",
    [
        "StateVec",
        "Qulacs",
        "CuStateVec",
        "MPS",
        "QuestStateVec",
    ],
)
def test_comp_basis_circ_and_measure(simulator: str) -> None:
    """Test computational basis circuit and measurement."""
    qc = QuantumCircuit()
    qc.append({"Init": {0, 1, 2, 3}})

    # Step 1
    qc.append({"X": {0, 2}})  # |0000> -> |1010>

    final_vector = pc.zeros(shape=(2**4,), dtype=pc.dtypes.complex128)
    final_vector[10] = 1  # |1010>

    # Run the circuit and compare results
    verify(simulator, qc, final_vector)

    # Insert detailed debug prints after verify
    sim_class = check_dependencies(simulator)
    sim_instance = sim_class(len(qc.qudits))
    sim_instance.run_circuit(qc)

    # Step 2
    qc.append({"CX": {(2, 1)}})  # |1010> -> |1110>

    final_vector = pc.zeros(shape=(2**4,), dtype=pc.dtypes.complex128)
    final_vector[14] = 1  # |1110>

    # Run the circuit and compare results for Step 2
    verify(simulator, qc, final_vector)
    sim_instance.run_circuit(qc)


@pytest.mark.parametrize(
    "simulator",
    [
        "StateVec",
        "Qulacs",
        "CuStateVec",
        "MPS",
        "QuestStateVec",
    ],
)
def test_all_gate_circ(simulator: str) -> None:
    """Test circuit with all quantum gates.

    Note:
        For MPS simulator, uses reduced bond dimension (chi=32) to limit computational
        cost while maintaining reasonable accuracy. MPS tests take longer due to gate
        application overhead in the tensor network backend.
    """
    # Use chi=32 for MPS to balance speed and accuracy
    # This limits bond dimension and speeds up the 4-qubit test
    sim_kwargs = {"chi": 32} if simulator == "MPS" else {}

    # Generate three different arbitrary states
    qcs: list[QuantumCircuit] = []
    qcs.append(generate_random_state(seed=1234))
    qcs.append(generate_random_state(seed=5555))
    qcs.append(generate_random_state(seed=42))

    # Verify that each of these states matches with StateVec
    for qc in qcs:
        compare_against_statevec(simulator, qc, **sim_kwargs)

    # Apply each gate on randomly generated states and compare again
    for qc in qcs:
        qc.append({"SZZ": {(3, 2)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"RX": {0, 2}}, angles=(pc.f64.frac_pi_4,))
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SXXdg": {(0, 3)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"RY": {0, 3}}, angles=(pc.f64.pi / 8,))
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"RZZ": {(0, 3)}}, angles=(pc.f64.pi / 16,))
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"RZ": {1, 3}}, angles=(pc.f64.pi / 16,))
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"R1XY": {2}}, angles=(pc.f64.pi / 16, pc.f64.frac_pi_2))
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"I": {0, 1, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"X": {1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Y": {2, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"CY": {(2, 3), (0, 1)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SYY": {(1, 2)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Z": {2, 0}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"H": {3, 1}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"RYY": {(2, 1)}}, angles=(pc.f64.pi / 8,))
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SZZdg": {(3, 1)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"F": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"CX": {(0, 1), (3, 2)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Fdg": {3, 1}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SYYdg": {(1, 3)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SX": {1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append(
            {"R2XXYYZZ": {(0, 3)}},
            angles=(pc.f64.frac_pi_4, pc.f64.pi / 16, pc.f64.frac_pi_2),
        )
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SY": {2, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SZ": {2, 0}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SZdg": {1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"CZ": {(1, 3)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SXdg": {2, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SYdg": {2, 0}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"T": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SXX": {(0, 2)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"SWAP": {(3, 0)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Tdg": {3, 1}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"RXX": {(1, 3)}}, angles=(pc.f64.frac_pi_4,))
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Q": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Qd": {0, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"R": {0}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Rd": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"S": {0, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"Sd": {0}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"H2": {2, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"H3": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"H4": {2, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"H5": {0, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"H6": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"F2": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"F2d": {0, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"F3": {2, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"F3d": {0, 1, 2}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"F4": {2, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"F4d": {0, 3}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"CNOT": {(0, 1)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"G": {(1, 3)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)
        qc.append({"II": {(3, 2)}})
        compare_against_statevec(simulator, qc, **sim_kwargs)

        # Measure
        qc.append({"Measure": {0, 1, 2, 3}})
        check_measurement(simulator, qc)


@pytest.mark.parametrize(
    "simulator",
    [
        "StateVec",
        "Qulacs",
        "CuStateVec",
        "QuestStateVec",
    ],
)
def test_hybrid_engine_no_noise(simulator: str) -> None:
    """Test that HybridEngine can use these simulators."""
    check_dependencies(simulator)

    n_shots = 1000
    phir_folder = Path(__file__).parent.parent / "phir"

    sim = HybridEngine(qsim=simulator)
    with (phir_folder / "bell_qparallel.phir.json").open() as f:
        program = json.load(f)
    results = sim.run(
        program=program,
        shots=n_shots,
        seed=42,
    )

    register = "c" if "c" in results else "m"
    result_values = results[register]
    assert pc.isclose(
        result_values.count("00") / n_shots,
        result_values.count("11") / n_shots,
        rtol=0.0,
        atol=0.1,
    )


@pytest.mark.parametrize(
    "simulator",
    [
        "StateVec",
        "Qulacs",
        "CuStateVec",
        "QuestStateVec",
    ],
)
def test_hybrid_engine_noisy(simulator: str) -> None:
    """Test that HybridEngine with noise can use these simulators."""
    check_dependencies(simulator)

    n_shots = 1000
    phir_folder = Path(__file__).parent.parent / "phir"

    generic_errors = GenericErrorModel(
        error_params={
            "p1": 2e-1,
            "p2": 2e-1,
            "p_meas": 2e-1,
            "p_init": 1e-1,
            "p1_error_model": {
                "X": 0.25,
                "Y": 0.25,
                "Z": 0.25,
                "L": 0.25,
            },
        },
    )
    sim = HybridEngine(qsim=simulator, error_model=generic_errors)
    with (phir_folder / "example1_no_wasm.phir.json").open() as f:
        program = json.load(f)
    sim.run(
        program=program,
        shots=n_shots,
    )



"""Test While loop support in IR generator."""

import pytest
from pecos.slr import CReg, For, Main, QReg, While
from pecos.slr.gen_codes.guppy.ir_generator import IRGuppyGenerator
from pecos.slr.misc import Comment
from pecos.slr.qeclib import qubit
from pecos.slr.qeclib.qubit.measures import Measure


def test_ir_while_loop_basic() -> None:
    """Test basic while loop generation."""
    prog = Main(
        q := QReg("q", 1),
        c := CReg("c", 1),
        # While loop with a simple condition
        # In real use, condition would come from measurement or input
        While(c[0]).Do(
            Comment("Apply operation in loop"),
            qubit.H(q[0]),
        ),
        Measure(q[0]) > c[0],
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check that while loop is generated
    assert "while " in code
    assert "# Apply operation in loop" in code


def test_ir_while_loop_with_quantum() -> None:
    """Test while loop with quantum operations."""
    prog = Main(
        q := QReg("q", 2),
        control := CReg("control", 1),
        result := CReg("result", 2),
        # Initialize with H on first qubit
        qubit.H(q[0]),
        # While loop controlled by classical bit
        While(control[0]).Do(
            Comment("Quantum operations in loop"),
            qubit.CX(q[0], q[1]),
            qubit.H(q[1]),
            Comment("In practice, control bit would be updated here"),
        ),
        # Measure both qubits
        Measure(q) > result,
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check structure
    assert "while " in code
    assert "quantum.cx" in code
    assert "quantum.h" in code
    assert "quantum.measure_array(q)" in code


def test_ir_nested_while_loops() -> None:
    """Test nested while loops."""
    prog = Main(
        q := QReg("q", 1),
        outer_cond := CReg("outer_cond", 1),
        inner_cond := CReg("inner_cond", 1),
        result := CReg("result", 1),
        # Outer while loop
        While(outer_cond[0]).Do(
            Comment("Outer loop"),
            # Inner while loop
            While(inner_cond[0]).Do(
                Comment("Inner loop"),
                qubit.X(q[0]),
            ),
            qubit.Z(q[0]),
        ),
        Measure(q[0]) > result[0],
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check for nested structure
    assert code.count("while ") >= 2  # At least 2 while statements
    assert "# Outer loop" in code
    assert "# Inner loop" in code


def test_ir_while_error_in_qasm() -> None:
    """Test that While loops raise error in QASM generator."""
    from pecos.slr.gen_codes.gen_qasm import QASMGenerator

    prog = Main(
        c := CReg("c", 1),
        While(c[0]).Do(
            Comment("This should fail in QASM"),
        ),
    )

    gen = QASMGenerator(_internal=True)

    # Should raise NotImplementedError
    with pytest.raises(NotImplementedError) as exc_info:
        gen.generate_block(prog)

    assert "While loops are not supported in QASM" in str(exc_info.value)


def test_ir_for_loop_placeholder() -> None:
    """Test For loop generates placeholder (not fully implemented yet)."""
    prog = Main(
        q := QReg("q", 3),
        c := CReg("c", 3),
        # For loop (generates TODO for now)
        For("i", 0, 3).Do(
            Comment("Apply H to each qubit"),
            qubit.H(q[0]),  # Would use i to index in full implementation
        ),
        # Measure all
        Measure(q) > c,
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # For loops should now be properly implemented
    assert "for i in range(0, 3):" in code


def test_while_loop_quantum_resource_handling() -> None:
    """Test that quantum resources are properly tracked in while loops."""
    prog = Main(
        q := QReg("q", 3),
        ancilla := QReg("ancilla", 1),
        cond := CReg("cond", 1),
        results := CReg("results", 4),
        # Initialize
        qubit.H(q[0]),
        qubit.H(q[1]),
        # While loop that uses quantum resources
        While(cond[0]).Do(
            Comment("Use ancilla in loop"),
            qubit.H(ancilla[0]),
            qubit.CX(q[0], ancilla[0]),
            Measure(ancilla[0]) > results[0],
            Comment("Ancilla is consumed each iteration"),
        ),
        # Measure remaining qubits
        Measure(q) > results[1:4],
    )

    gen = IRGuppyGenerator(_internal=True)
    gen.generate_block(prog)
    code = gen.get_output()

    # Check that measurements are properly handled
    assert "while " in code
    # With dynamic allocation, ancilla is allocated locally in the loop
    assert "ancilla_0_local = quantum.qubit()" in code
    assert "quantum.measure(ancilla_0_local)" in code
    assert "quantum.measure_array(q)" in code



"""Comprehensive testing of Guppy language features across both HUGR-LLVM and PHIR pipelines.

This test suite systematically validates that both compilation pipelines can handle
the full spectrum of Guppy language capabilities, from basic quantum operations
to advanced classical-quantum hybrid programs.
"""

from typing import TYPE_CHECKING, Any

import pytest
from guppylang import guppy
from guppylang.std.quantum import cx, h, measure, qubit, x, y, z
from pecos import Guppy, sim
from pecos_rslib import state_vector

if TYPE_CHECKING:
    from pecos.protocols import GuppyCallable


def decode_integer_results(results: list[int], n_bits: int) -> list[tuple[bool, ...]]:
    """Decode integer-encoded results back to tuples of booleans.

    When guppy functions return tuples of bools, sim encodes them
    as integers where bit i represents the i-th boolean in the tuple.
    """
    decoded = []
    for val in results:
        bits = [bool(val & (1 << i)) for i in range(n_bits)]
        decoded.append(tuple(bits))
    return decoded


def get_guppy_backends() -> dict[str, Any]:
    """Get available backends."""
    return {"guppy_available": True, "rust_backend": True}


class GuppyPipelineTest:
    """Helper class for testing Guppy programs on both pipelines."""

    def __init__(self) -> None:
        """Initialize test helper with available backends."""
        self.backends = get_guppy_backends()

    def test_function_on_both_pipelines(
        self,
        func: "GuppyCallable",
        shots: int = 10,
        seed: int = 42,
        **kwargs: object,
    ) -> dict[str, Any]:
        """Test a Guppy function (using the Rust backend)."""
        results = {}

        # Test with Rust backend (the only backend)
        if self.backends.get("rust_backend", False):
            try:
                # Use sim() API instead of run_guppy
                n_qubits = kwargs.get("n_qubits", kwargs.get("max_qubits", 10))
                builder = sim(Guppy(func)).qubits(n_qubits).quantum(state_vector())
                if seed is not None:
                    builder = builder.seed(seed)
                result_obj = builder.run(shots)
                result_dict = result_obj.to_dict()

                # Format results to match expected structure
                measurements = []
                if "measurements" in result_dict:
                    # measurements is a list of lists like [[1], [0, 1], ...]
                    # For functions returning single bool, extract the last measurement per shot
                    raw_measurements = result_dict["measurements"]
                    if raw_measurements and isinstance(raw_measurements[0], list):
                        # Check if function returns single bool or tuple
                        import inspect

                        actual_func = func
                        if hasattr(func, "wrapped") and hasattr(
                            func.wrapped,
                            "python_func",
                        ):
                            actual_func = func.wrapped.python_func
                        try:
                            sig = inspect.signature(actual_func)
                            return_type = sig.return_annotation
                            is_tuple_return = hasattr(return_type, "__origin__") and return_type.__origin__ is tuple
                        except (ValueError, TypeError):
                            is_tuple_return = False

                        if is_tuple_return:
                            # Return full measurement tuples
                            measurements = [tuple(m) for m in raw_measurements]
                        else:
                            # For single bool return, take the last measurement from each shot
                            measurements = [m[-1] if m else 0 for m in raw_measurements]
                    else:
                        measurements = raw_measurements
                elif "measurement_0" in result_dict:
                    # Handle multiple measurements
                    num_shots = len(result_dict["measurement_0"])
                    measurement_keys = sorted(
                        [k for k in result_dict if k.startswith("measurement_")],
                    )
                    num_measurements = len(measurement_keys)

                    for i in range(num_shots):
                        result_tuple = [bool(result_dict[key][i]) for key in measurement_keys]

                        # Check function signature to determine if it returns a tuple
                        # For now, if there's more than one measurement but function returns single bool,
                        # take the last measurement as the return value
                        import inspect

                        # For Guppy functions, we need to check the wrapped function
                        actual_func = func
                        if hasattr(func, "wrapped") and hasattr(
                            func.wrapped,
                            "python_func",
                        ):
                            actual_func = func.wrapped.python_func

                        sig = inspect.signature(actual_func)
                        return_type = sig.return_annotation

                        # Check if return type is a tuple
                        is_tuple_return = hasattr(return_type, "__origin__") and return_type.__origin__ is tuple
                        if is_tuple_return or num_measurements == 1:
                            # For tuple returns or single measurement, use all measurements
                            measurements.append(
                                (tuple(result_tuple) if len(result_tuple) > 1 else result_tuple[0]),
                            )
                        else:
                            # For single bool return with multiple measurements, take the last one
                            measurements.append(result_tuple[-1])
                elif "result" in result_dict:
                    measurements = result_dict["result"]

                func_name = getattr(
                    func,
                    "__name__",
                    getattr(func, "name", "quantum_func"),
                )
                result = {
                    "results": measurements,
                    "shots": shots,
                    "function_name": func_name,
                }
                results["hugr_llvm"] = {
                    "success": True,
                    "result": result,
                    "error": None,
                }
            except Exception as e:
                results["hugr_llvm"] = {
                    "success": False,
                    "result": None,
                    "error": str(e),
                }

        return results


@pytest.fixture
def pipeline_tester() -> GuppyPipelineTest:
    """Fixture providing the pipeline testing helper."""
    import gc

    # Force garbage collection to clean up any lingering resources
    gc.collect()

    # Create fresh test instance
    tester = GuppyPipelineTest()

    yield tester

    # Force garbage collection to clean up test resources
    gc.collect()


# ============================================================================
# BASIC QUANTUM OPERATIONS TESTS
# ============================================================================


class TestBasicQuantumOperations:
    """Test basic quantum gate operations on both pipelines."""

    def test_single_qubit_hadamard(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test Hadamard gate on single qubit."""

        @guppy
        def hadamard_test() -> bool:
            q = qubit()
            h(q)
            return measure(q)

        results = pipeline_tester.test_function_on_both_pipelines(
            hadamard_test,
            shots=50,
        )
        assert results.get("hugr_llvm", {}).get(
            "success",
            False,
        ), f"HUGR-LLVM failed: {results.get('hugr_llvm', {}).get('error')}"
        # PHIR might not be available on all systems
        if "phir" in results:
            # print(f"PHIR result: {results['phir']}")
            pass

    def test_pauli_gates(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test all Pauli gates (X, Y, Z)."""

        @guppy
        def pauli_x_test() -> bool:
            q = qubit()
            x(q)  # Should flip |0⟩ to |1⟩
            return measure(q)

        @guppy
        def pauli_y_test() -> bool:
            q = qubit()
            y(q)  # Should flip |0⟩ to |1⟩ with phase
            return measure(q)

        @guppy
        def pauli_z_test() -> bool:
            q = qubit()
            z(q)  # Should leave |0⟩ unchanged
            return measure(q)

        # Test X gate - should measure |1⟩ deterministically with fixed seed
        results_x = pipeline_tester.test_function_on_both_pipelines(
            pauli_x_test,
            shots=100,
            seed=42,
        )
        if results_x.get("hugr_llvm", {}).get("success"):
            ones_count = sum(results_x["hugr_llvm"]["result"]["results"])
            # X gate should flip |0⟩ to |1⟩, expect 100% ones
            assert ones_count == 100, f"X gate should produce all 1s, got {ones_count}/100"

        # Test Y gate - should measure |1⟩ deterministically
        results_y = pipeline_tester.test_function_on_both_pipelines(
            pauli_y_test,
            shots=100,
            seed=42,
        )
        if results_y.get("hugr_llvm", {}).get("success"):
            ones_count = sum(results_y["hugr_llvm"]["result"]["results"])
            # Y gate should flip |0⟩ to |1⟩ with phase, expect 100% ones
            assert ones_count == 100, f"Y gate should produce all 1s, got {ones_count}/100"

        # Test Z gate - should measure |0⟩ deterministically
        results_z = pipeline_tester.test_function_on_both_pipelines(
            pauli_z_test,
            shots=100,
            seed=42,
        )
        if results_z.get("hugr_llvm", {}).get("success"):
            ones_count = sum(results_z["hugr_llvm"]["result"]["results"])
            # Z gate should leave |0⟩ unchanged, expect 0% ones
            assert ones_count == 0, f"Z gate should produce all 0s, got {ones_count}/100"

    def test_bell_state_entanglement(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test Bell state creation and entanglement."""

        @guppy
        def bell_state() -> tuple[bool, bool]:
            q0, q1 = qubit(), qubit()
            h(q0)
            cx(q0, q1)
            return measure(q0), measure(q1)

        results = pipeline_tester.test_function_on_both_pipelines(bell_state, shots=50)

        # Verify HUGR-LLVM pipeline results
        if results.get("hugr_llvm", {}).get("success"):
            measurements = results["hugr_llvm"]["result"]["results"]
            # Check if measurements are already tuples or need decoding
            if measurements and isinstance(measurements[0], tuple):
                # Already decoded as tuples
                decoded_measurements = measurements
            else:
                # Decode integer-encoded results
                decoded_measurements = decode_integer_results(measurements, 2)
            correlated = sum(1 for (a, b) in decoded_measurements if a == b)
            correlation_rate = correlated / len(decoded_measurements)
            assert correlation_rate > 0.8, f"Bell state should be highly correlated, got {correlation_rate:.2%}"

        # Verify PHIR pipeline results if available
        if results.get("phir", {}).get("success"):
            measurements = results["phir"]["result"]["results"]
            # Decode integer-encoded results
            decoded_measurements = decode_integer_results(measurements, 2)
            correlated = sum(1 for (a, b) in decoded_measurements if a == b)
            correlation_rate = correlated / len(decoded_measurements)
            assert correlation_rate > 0.8, f"PHIR Bell state should be highly correlated, got {correlation_rate:.2%}"


# ============================================================================
# CLASSICAL COMPUTATION TESTS
# ============================================================================


class TestClassicalComputation:
    """Test classical computation capabilities in both pipelines."""

    def test_boolean_operations(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test boolean logic operations."""

        @guppy
        def boolean_and_test() -> bool:
            # Simple boolean logic with quantum measurement
            q = qubit()
            result = measure(q)  # Will be False (|0⟩)
            return result and True

        @guppy
        def boolean_or_test() -> bool:
            q = qubit()
            x(q)  # Flip to |1⟩
            result = measure(q)  # Will be True
            return result or False

        # Test AND operation
        pipeline_tester.test_function_on_both_pipelines(
            boolean_and_test,
            shots=10,
        )

        # Test OR operation
        pipeline_tester.test_function_on_both_pipelines(
            boolean_or_test,
            shots=10,
        )

    def test_classical_arithmetic(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test basic arithmetic operations."""

        # NOTE: This may fail on current pipelines due to limited classical support
        @guppy
        def arithmetic_test() -> int:
            # Simple arithmetic that doesn't depend on quantum measurements
            a = 5
            b = 3
            return a + b

        results = pipeline_tester.test_function_on_both_pipelines(
            arithmetic_test,
            shots=5,
        )

        # Document current limitations
        if not results.get("hugr_llvm", {}).get("success"):
            pass
        if not results.get("phir", {}).get("success"):
            pass


# ============================================================================
# HYBRID QUANTUM-CLASSICAL TESTS
# ============================================================================


class TestHybridPrograms:
    """Test hybrid quantum-classical programs."""

    def test_conditional_quantum_operations(
        self,
        pipeline_tester: GuppyPipelineTest,
    ) -> None:
        """Test quantum operations conditional on classical results."""

        @guppy
        def conditional_gate() -> bool:
            q1 = qubit()
            q2 = qubit()

            # Measure first qubit
            result1 = measure(q1)  # Will be False (|0⟩)

            # Apply gate to second qubit based on first measurement
            if result1:
                x(q2)  # This won't execute since result1 is False

            return measure(q2)  # Should be False

        results = pipeline_tester.test_function_on_both_pipelines(
            conditional_gate,
            shots=20,
        )
        if results.get("hugr_llvm", {}).get("success"):
            measurements = results["hugr_llvm"]["result"]["results"]
            # Results are boolean values, count True values
            sum(1 for r in measurements if r)
            # When HUGR to LLVM compilation is properly implemented,
            # this should assert:
            # assert ones_count < 5, f"Conditional gate failed, got {ones_count}/20 ones"

    def test_measurement_feedback(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test feedback based on mid-circuit measurements."""

        @guppy
        def feedback_circuit() -> tuple[bool, bool]:
            q1 = qubit()
            q2 = qubit()

            # Create superposition on first qubit
            h(q1)
            result1 = measure(q1)

            # Apply correction to second qubit based on measurement
            if result1:
                x(q2)  # Flip second qubit if first was |1⟩

            return result1, measure(q2)

        pipeline_tester.test_function_on_both_pipelines(
            feedback_circuit,
            shots=50,
        )


# ============================================================================
# ADVANCED QUANTUM ALGORITHMS (PLACEHOLDER)
# ============================================================================


class TestAdvancedAlgorithms:
    """Test advanced quantum algorithms (to be implemented)."""

    def test_quantum_fourier_transform(
        self,
        pipeline_tester: GuppyPipelineTest,
    ) -> None:
        """Test quantum Fourier transform on 2 qubits."""
        from guppylang.std.angles import pi
        from guppylang.std.quantum import crz, cx, h, measure, qubit, x

        @guppy
        def qft_2qubit() -> tuple[bool, bool]:
            """2-qubit QFT implementation."""
            # Initialize qubits
            q0 = qubit()
            q1 = qubit()

            # Apply X to q1 to create input state |01⟩
            x(q1)

            # QFT circuit for 2 qubits
            # First qubit
            h(q0)
            # Controlled rotation
            # In QFT, we use controlled-R_2 which is a phase rotation by π/2
            # We can implement this using CRZ
            crz(q1, q0, pi / 2)

            # Second qubit
            h(q1)

            # Swap qubits (using 3 CNOTs since we don't have swap)
            cx(q0, q1)
            cx(q1, q0)
            cx(q0, q1)

            # Measure
            return measure(q0), measure(q1)

        results = pipeline_tester.test_function_on_both_pipelines(qft_2qubit, shots=100)

        if results.get("hugr_llvm", {}).get("success"):
            # QFT of |01⟩ should give a specific pattern
            measurements = results["hugr_llvm"]["result"]["results"]
            # print(f"QFT results distribution: {set(measurements)}")
            # The test passes if we get results without errors
            assert len(measurements) == 100

    def test_deutsch_josza_algorithm(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test Deutsch-Josza algorithm for 2-bit function."""
        from guppylang.std.quantum import cx, h, measure, qubit, x

        @guppy
        def deutsch_josza_constant() -> tuple[bool, bool]:
            """Deutsch-Josza algorithm with constant oracle (f(x)=0)."""
            # Initialize qubits
            q0 = qubit()  # First input qubit
            q1 = qubit()  # Second input qubit
            anc = qubit()  # Ancilla qubit

            # Prepare ancilla in |1⟩ and apply H to get |->⟩
            x(anc)
            h(anc)

            # Apply H to input qubits
            h(q0)
            h(q1)

            # Oracle for constant function f(x) = 0
            # Does nothing since f(x) = 0 for all x

            # Apply H to input qubits again
            h(q0)
            h(q1)

            # Measure input qubits (ancilla can be discarded)
            return measure(q0), measure(q1)

        @guppy
        def deutsch_josza_balanced() -> tuple[bool, bool]:
            """Deutsch-Josza algorithm with balanced oracle."""
            # Initialize qubits
            q0 = qubit()  # First input qubit
            q1 = qubit()  # Second input qubit
            anc = qubit()  # Ancilla qubit

            # Prepare ancilla in |->⟩
            x(anc)
            h(anc)

            # Apply H to input qubits
            h(q0)
            h(q1)

            # Oracle for balanced function: f(00)=0, f(01)=1, f(10)=1, f(11)=0
            # This is implemented using controlled operations
            cx(q1, anc)  # Flip ancilla if q1 is |1⟩
            cx(q0, anc)  # Flip ancilla if q0 is |1⟩

            # Apply H to input qubits again
            h(q0)
            h(q1)

            # Measure input qubits
            return measure(q0), measure(q1)

        # Test constant function
        results_const = pipeline_tester.test_function_on_both_pipelines(
            deutsch_josza_constant,
            shots=100,
        )
        if results_const.get("hugr_llvm", {}).get("success"):
            measurements = results_const["hugr_llvm"]["result"]["results"]
            # Decode integer-encoded results
            decoded_measurements = decode_integer_results(measurements, 2)
            # For constant function, should measure |00⟩ with high probability
            zeros = sum(1 for (a, b) in decoded_measurements if not a and not b)
            assert zeros > 95, f"Constant oracle should give |00⟩, got {zeros}/100"

        # Test balanced function
        results_bal = pipeline_tester.test_function_on_both_pipelines(
            deutsch_josza_balanced,
            shots=100,
        )
        if results_bal.get("hugr_llvm", {}).get("success"):
            measurements = results_bal["hugr_llvm"]["result"]["results"]
            # Decode integer-encoded results
            decoded_measurements = decode_integer_results(measurements, 2)
            # For balanced function, should never measure |00⟩
            zeros = sum(1 for (a, b) in decoded_measurements if not a and not b)
            assert zeros < 5, f"Balanced oracle should not give |00⟩, got {zeros}/100"

    def test_grover_search(self, pipeline_tester: GuppyPipelineTest) -> None:
        """Test Grover's search algorithm for 2-qubit search space."""
        from guppylang.std.quantum import cz, h, measure, qubit, x

        @guppy
        def grover_2qubit() -> tuple[bool, bool]:
            """Grover's algorithm searching for |11⟩ in 2-qubit space."""
            # Initialize qubits
            q0 = qubit()
            q1 = qubit()

            # Initialize in uniform superposition
            h(q0)
            h(q1)

            # Grover iteration (just 1 iteration for 2 qubits)
            # Oracle: mark |11⟩ state
            # We use CZ which adds a phase to |11⟩
            cz(q0, q1)

            # Diffusion operator (inversion about average)
            # Apply H gates
            h(q0)
            h(q1)

            # Apply X gates
            x(q0)
            x(q1)

            # Apply CZ (multi-controlled Z, but for 2 qubits just CZ)
            cz(q0, q1)

            # Apply X gates
            x(q0)
            x(q1)

            # Apply H gates
            h(q0)
            h(q1)

            # Measure
            return measure(q0), measure(q1)

        results = pipeline_tester.test_function_on_both_pipelines(
            grover_2qubit,
            shots=100,
        )
        if results.get("hugr_llvm", {}).get("success"):
            measurements = results["hugr_llvm"]["result"]["results"]
            # Check if measurements are already tuples or need decoding
            if measurements and isinstance(measurements[0], tuple):
                # Already decoded as tuples
                decoded_measurements = measurements
            else:
                # Decode integer-encoded results
                decoded_measurements = decode_integer_results(measurements, 2)
            # Should find |11⟩ with high probability after 1 Grover iteration
            found = sum(1 for (a, b) in decoded_measurements if a and b)
            assert found > 70, f"Grover should amplify |11⟩, got {found}/100"



# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Surface code quantum error correction library."""

from pecos.slr.qeclib.surface.gate_sets.surface_std_gates import SurfaceStdGates
from pecos.slr.qeclib.surface.layouts.layout_base import LatticeType
from pecos.slr.qeclib.surface.patch_builders import SurfacePatchBuilder
from pecos.slr.qeclib.surface.patches.patch_base import SurfacePatchOrientation
from pecos.slr.qeclib.surface.patches.surface_patches import (
    NonRotatedSurfacePatch,
    RotatedSurfacePatch,
)
from pecos.slr.qeclib.surface.visualization.lattice_2d import (
    Lattice2DConfig,
    Lattice2DView,
)

__all__ = [
    "Lattice2DConfig",
    "Lattice2DView",
    "LatticeType",
    "NonRotatedSurfacePatch",
    "RotatedSurfacePatch",
    "SurfacePatchBuilder",
    "SurfacePatchOrientation",
    "SurfaceStdGates",
]



"""Pytest fixtures for SLR tests."""

import pytest
from pecos.slr import CReg, Main, Permute, QReg
from pecos.slr.qeclib import qubit


@pytest.fixture
def basic_permutation_program() -> tuple:
    """Create a basic program with permutation of classical registers."""
    a = CReg("a", 2)
    b = CReg("b", 2)

    prog = Main(
        a,
        b,
        Permute(
            [a[0], b[1]],
            [b[1], a[0]],
        ),
        a[0].set(1),  # Should become b[1] = 1 after permutation
    )

    return prog, a, b


@pytest.fixture
def same_register_permutation_program() -> tuple:
    """Create a program with permutation within the same register."""
    a = CReg("a", 3)

    prog = Main(
        a,
        Permute(
            [a[0], a[1], a[2]],
            [a[2], a[0], a[1]],
        ),
        a[0].set(1),  # Should become a[2] = 1
        a[1].set(0),  # Should become a[0] = 0
        a[2].set(1),  # Should become a[1] = 1
    )

    return prog, a


@pytest.fixture
def quantum_permutation_program() -> tuple:
    """Create a program with permutation of quantum registers."""
    a = QReg("a", 2)
    b = QReg("b", 2)

    prog = Main(
        a,
        b,
        Permute(
            [a[0], b[0]],
            [b[0], a[0]],
        ),
        qubit.H(a[0]),  # Should become H(b[0]) after permutation
        qubit.CX(a[0], a[1]),  # Should become CX(b[0], a[1]) after permutation
    )

    return prog, a, b


@pytest.fixture
def measurement_program() -> tuple:
    """Create a program with permutation and measurements."""
    a = QReg("a", 2)
    b = QReg("b", 2)
    m = CReg("m", 2)
    n = CReg("n", 2)

    prog = Main(
        a,
        b,
        m,
        n,
        # Apply permutations to both quantum and classical registers
        Permute(
            [a[0], b[0]],
            [b[0], a[0]],
        ),
        Permute(
            [m[0], n[0]],
            [n[0], m[0]],
        ),
        # Apply quantum operations
        qubit.H(a[0]),
        qubit.CX(a[0], b[0]),
    )

    return prog, a, b, m, n


@pytest.fixture
def individual_measurement_program() -> tuple:
    """Create a program with permutation and individual measurements."""
    a = QReg("a", 2)
    b = QReg("b", 2)
    m = CReg("m", 2)
    n = CReg("n", 2)

    prog = Main(
        a,
        b,
        m,
        n,
        # Apply permutations to both quantum and classical registers
        Permute(
            [a[0], b[0]],
            [b[0], a[0]],
        ),
        Permute(
            [m[0], n[0]],
            [n[0], m[0]],
        ),
        # Apply quantum operations
        qubit.H(a[0]),
        qubit.CX(a[0], b[0]),
        # Add individual measurements
        qubit.Measure(a[0]) > m[0],
        qubit.Measure(a[1]) > m[1],
    )

    return prog, a, b, m, n


@pytest.fixture
def register_measurement_program() -> tuple:
    """Create a program with permutation and register-wide measurements."""
    a = QReg("a", 2)
    b = QReg("b", 2)
    m = CReg("m", 2)
    n = CReg("n", 2)

    prog = Main(
        a,
        b,
        m,
        n,
        # Apply permutations to both quantum and classical registers
        Permute(
            [a[0], b[0]],
            [b[0], a[0]],
        ),
        Permute(
            [m[0], n[0]],
            [n[0], m[0]],
        ),
        # Apply quantum operations
        qubit.H(a[0]),
        qubit.CX(a[0], b[0]),
        # Add register-wide measurement
        qubit.Measure(a) > m,
    )

    return prog, a, b, m, n



# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Gate bindings for cuStateVec quantum simulator.

This module provides gate operation bindings for the NVIDIA cuStateVec-based quantum simulator, organizing and
exposing GPU-accelerated quantum gate implementations for high-performance quantum circuit simulation.
"""

import pecos.simulators.custatevec.gates_one_qubit as one_q
import pecos.simulators.custatevec.gates_two_qubit as two_q
from pecos.simulators.custatevec.gates_init import init_one, init_zero
from pecos.simulators.custatevec.gates_meas import meas_z

# Supporting gates from table:
#   https://github.com/Quantinuum/phir/blob/main/spec.md#table-ii---quantum-operations

gate_dict = {
    "Init": init_zero,
    "Init +Z": init_zero,
    "Init -Z": init_one,
    "init |0>": init_zero,
    "init |1>": init_one,
    "leak": init_zero,
    "leak |0>": init_zero,
    "leak |1>": init_one,
    "unleak |0>": init_zero,
    "unleak |1>": init_one,
    "Measure": meas_z,
    "measure Z": meas_z,
    "I": one_q.identity,
    "X": one_q.X,
    "Y": one_q.Y,
    "Z": one_q.Z,
    "RX": one_q.RX,
    "RY": one_q.RY,
    "RZ": one_q.RZ,
    "R1XY": one_q.R1XY,
    "RXY1Q": one_q.R1XY,
    "SX": one_q.SX,
    "SXdg": one_q.SXdg,
    "SqrtX": one_q.SX,
    "SqrtXd": one_q.SXdg,
    "SY": one_q.SY,
    "SYdg": one_q.SYdg,
    "SqrtY": one_q.SY,
    "SqrtYd": one_q.SYdg,
    "SZ": one_q.SZ,
    "SZdg": one_q.SZdg,
    "SqrtZ": one_q.SZ,
    "SqrtZd": one_q.SZdg,
    "H": one_q.H,
    "F": one_q.F,
    "Fdg": one_q.Fdg,
    "T": one_q.T,
    "Tdg": one_q.Tdg,
    "CX": two_q.CX,
    "CY": two_q.CY,
    "CZ": two_q.CZ,
    "RXX": two_q.RXX,
    "RYY": two_q.RYY,
    "RZZ": two_q.RZZ,
    "R2XXYYZZ": two_q.R2XXYYZZ,
    "SXX": two_q.SXX,
    "SXXdg": two_q.SXXdg,
    "SYY": two_q.SYY,
    "SYYdg": two_q.SYYdg,
    "SZZ": two_q.SZZ,
    "SqrtZZ": two_q.SZZ,
    "SZZdg": two_q.SZZdg,
    "SWAP": two_q.SWAP,
    "Q": one_q.SX,
    "Qd": one_q.SXdg,
    "R": one_q.SY,
    "Rd": one_q.SYdg,
    "S": one_q.SZ,
    "Sd": one_q.SZdg,
    "H1": one_q.H,
    "H2": one_q.H2,
    "H3": one_q.H3,
    "H4": one_q.H4,
    "H5": one_q.H5,
    "H6": one_q.H6,
    "H+z+x": one_q.H,
    "H-z-x": one_q.H2,
    "H+y-z": one_q.H3,
    "H-y-z": one_q.H4,
    "H-x+y": one_q.H5,
    "H-x-y": one_q.H6,
    "F1": one_q.F,
    "F1d": one_q.Fdg,
    "F2": one_q.F2,
    "F2d": one_q.F2d,
    "F3": one_q.F3,
    "F3d": one_q.F3d,
    "F4": one_q.F4,
    "F4d": one_q.F4d,
    "CNOT": two_q.CX,
    "G": two_q.G,
    "II": one_q.identity,
}



# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Threshold curve analysis and fitting for quantum error correction.

This module provides utilities for analyzing and fitting threshold curves
in quantum error correction, including error rate scaling and critical
threshold determination.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pecos as pc

if TYPE_CHECKING:
    from collections.abc import Callable

    from pecos import (
        Array,
        f64,
    )


def func(
    x: tuple[Array[f64], Array[f64]],
    pth: float,
    v0: float,
    a: float,
    b: float,
    c: float,
) -> float | Array[f64]:
    """Fit error rates to determine threshold using polynomial expansion.

    Function that represents the curve to fit error rates to in order to determine the threshold. (see:
    arXiv:quant-ph/0207088).

    Probabilities are fine as long as p > 1/(4*distance). See paper by Watson and Barrett (arXiv:1312.5213).

    Args:
    ----
        x: Tuple of (p, dist) where p is the physical error rate and dist is the code distance.
        pth: Threshold error rate parameter to be fitted.
        v0: Critical exponent parameter for scaling behavior.
        a: Constant term coefficient in the polynomial expansion.
        b: Linear term coefficient in the polynomial expansion.
        c: Quadratic term coefficient in the polynomial expansion.

    """
    p, dist = x

    x = (p - pth) * pc.power(dist, 1.0 / v0)

    return a + b * x + c * pc.power(x, 2)


def func2(
    x: tuple[Array[f64], Array[f64]],
    pth: float,
    v0: float,
    a: float,
    b: float,
    c: float,
    d: float,
    u: float,
) -> float | Array[f64]:
    """Fit error rates with finite-size correction to determine threshold.

    Function that represents the curve to fit error rates to in order to determine the threshold. (see:
    arXiv:quant-ph/0207088).

    Probabilities are fine as long as p > 1/(4*distance). See paper by Watson and Barrett (arXiv:1312.5213).

    Args:
    ----
        x: Tuple of (p, dist) where p is the physical error rate and dist is the code distance.
        pth: Threshold error rate parameter to be fitted.
        v0: Critical exponent parameter for scaling behavior.
        a: Constant term coefficient in the polynomial expansion.
        b: Linear term coefficient in the polynomial expansion.
        c: Quadratic term coefficient in the polynomial expansion.
        d: Coefficient for the finite-size correction term.
        u: Exponent parameter for the finite-size correction term.

    """
    p, dist = x

    x = (p - pth) * pc.power(dist, 1.0 / v0)

    z = a + b * x + c * pc.power(x, 2)

    z += d * pc.power(dist, -1.0 / u)

    return z


def func3(
    x: tuple[Array[f64], Array[f64]],
    pth: float,
    v0: float,
    a: float,
    b: float,
    c: float,
    d: float,
    uodd: float,
    ueven: float,
) -> float | Array[f64]:
    """Fit error rates with odd/even distance corrections to determine threshold.

    Function that represents the curve to fit error rates to in order to determine the threshold. (see:
    arXiv:quant-ph/0207088).

    Probabilities are fine as long as p > 1/(4*distance). See paper by Watson and Barrett (arXiv:1312.5213).

    Args:
    ----
        x: Tuple of (p, dist) where p is the physical error rate and dist is the code distance.
        pth: Threshold error rate parameter to be fitted.
        v0: Critical exponent parameter for scaling behavior.
        a: Constant term coefficient in the polynomial expansion.
        b: Linear term coefficient in the polynomial expansion.
        c: Quadratic term coefficient in the polynomial expansion.
        d: Coefficient for the finite-size correction term.
        uodd: Exponent parameter for finite-size corrections at odd distances.
        ueven: Exponent parameter for finite-size corrections at even distances.

    """
    p, dist = x

    x = (p - pth) * pc.power(dist, 1.0 / v0)

    z = pc.where(
        bool(dist % 2),
        d * pc.power(dist, -1.0 / uodd),
        d * pc.power(dist, -1.0 / ueven),
    )

    z += a + b * x + c * pc.power(x, 2)

    return z


def func4(
    x: tuple[Array[f64], Array[f64]],
    pth: float,
    v0: float,
    a: float,
    b: float,
) -> float | Array[f64]:
    """Fit error rates using exponential decay to determine threshold.

    Function that represents the curve to fit error rates to in order to determine the threshold. (see:
    arXiv:quant-ph/0207088).

    Probabilities are fine as long as p > 1/(4*distance). See paper by Watson and Barrett (arXiv:1312.5213).

    Args:
    ----
        x: Tuple of (p, dist) where p is the physical error rate and dist is the code distance.
        pth: Threshold error rate parameter to be fitted.
        v0: Critical exponent parameter for scaling behavior.
        a: Amplitude coefficient for the exponential decay.
        b: Decay rate coefficient in the exponential function.

    """
    p, dist = x

    x = (p - pth) * pc.power(dist, 1.0 / v0)

    return a * pc.exp(-b * pc.power(x, v0))


def func5(
    x: tuple[Array[f64], Array[f64]],
    pth: float,
    v0: float,
    a: float,
    b: float,
    c: float,
    d: float,
) -> float | Array[f64]:
    """Fit error rates using cubic polynomial to determine threshold.

    Function that represents the curve to fit error rates to in order to determine the threshold. (see:
    arXiv:quant-ph/0207088).

    Probabilities are fine as long as p > 1/(4*distance). See paper by Watson and Barrett (arXiv:1312.5213).

    Args:
    ----
        x: Tuple of (p, dist) where p is the physical error rate and dist is the code distance.
        pth: Threshold error rate parameter to be fitted.
        v0: Critical exponent parameter for scaling behavior.
        a: Constant term coefficient in the polynomial expansion.
        b: Linear term coefficient in the polynomial expansion.
        c: Quadratic term coefficient in the polynomial expansion.
        d: Cubic term coefficient in the polynomial expansion.

    """
    p, dist = x

    x = (p - pth) * pc.power(dist, 1.0 / v0)

    return a + b * x + c * pc.power(x, 2) + d * pc.power(x, 3)


def func6(
    x: tuple[Array[f64], Array[f64]],
    a: float,
    pth: float,
) -> float | Array[f64]:
    """Fit error rates using power law relationship to determine threshold.

    Function that represents the curve to fit error rates to in order to determine the threshold. (see:
    arXiv:quant-ph/0207088).

    Probabilities are fine as long as p > 1/(4*distance). See paper by Watson and Barrett (arXiv:1312.5213).

    Args:
    ----
        x: Tuple of (p, dist) where p is the physical error rate and dist is the code distance.
        a: Amplitude coefficient for the power law relationship.
        pth: Threshold error rate parameter.

    """
    p, dist = x

    return a * pc.power(p / pth, dist / 2)


def threshold_fit(
    plist: Array[f64] | list[float],
    dlist: Array[f64] | list[float],
    plog: Array[f64] | list[float],
    func: Callable[..., float | Array[f64]],
    p0: Array[f64] | list[float],
    maxfev: int = 100000,
    **kwargs: float | bool | str | None,
) -> tuple[Array[f64], Array[f64]]:
    """Fit threshold curve to logical error rate data.

    Args:
    ----
        plist: List of ps.
        dlist: List of distances.
        plog: List of logical error rates.
        func: Function to fit to.
        p0: Initial guess for the parameters.
        maxfev: Maximum number of function evaluations.
        **kwargs: Additional keyword arguments passed to curve_fit.

    """
    popt, pcov = pc.curve_fit(func, (plist, dlist), plog, p0, maxfev=maxfev, **kwargs)

    var = pc.diag(pcov)
    stdev = pc.sqrt(var)

    return popt, stdev


def _jackknife_threshold_core(
    plist: Array[f64] | list[float],
    dlist: Array[f64] | list[float],
    plog: Array[f64] | list[float],
    func: Callable[..., float | Array[f64]],
    p0: Array[f64] | list[float],
    maxfev: int,
    resample_indices: list[list[int]],
    *,
    verbose: bool = True,
    verbose_labels: list[str] | None = None,
) -> tuple[Array[f64], Array[f64]]:
    """Core jackknife resampling implementation for threshold fitting.

    Args:
        plist: List of probability values.
        dlist: List of distance values.
        plog: List of logical error probabilities.
        func: Fitting function to use.
        p0: Initial parameter guess.
        maxfev: Maximum function evaluations.
        resample_indices: List of index lists, each specifying which indices to include in that resample.
        verbose: If True, print progress information.
        verbose_labels: Optional labels for verbose output.

    Returns:
        Tuple of (mean_parameters, std_parameters).
    """
    opt_list = []

    for i, indices in enumerate(resample_indices):
        p_copy = plist[indices]
        plog_copy = plog[indices]
        dlist_copy = dlist[indices]

        result = threshold_fit(p_copy, dlist_copy, plog_copy, func, p0, maxfev)
        opt_list.append(result[0].tolist())

        if verbose and verbose_labels:
            print(verbose_labels[i])
            print("parameter values:", result[0])
            print(f"parameter stds: {result[1]}\n")

    # Convert to PECOS array for jackknife_stats_axis
    opt_array = pc.array(opt_list)

    # Use pecos-num jackknife_stats_axis to compute stats for all parameters at once
    # axis=0 means compute stats down columns (each column is a parameter)
    means, stds = pc.stats.jackknife_stats_axis(opt_array, axis=0)

    print(f"Mean: {means}")
    print(f"Std: {stds}")

    return means, stds


def jackknife_pd(
    plist: Array[f64] | list[float],
    dlist: Array[f64] | list[float],
    plog: Array[f64] | list[float],
    func: Callable[..., float | Array[f64]],
    p0: Array[f64] | list[float],
    maxfev: int = 100000,
    *,
    verbose: bool = True,
) -> tuple[Array[f64], Array[f64]]:
    """Perform jackknife resampling for parameter and distance data.

    Uses leave-one-out resampling where each data point (p, d, plog) is removed in turn.

    Args:
        plist: List of probability values.
        dlist: List of distance values.
        plog: List of logical error probabilities.
        func: Fitting function to use.
        p0: Initial parameter guess.
        maxfev: Maximum function evaluations.
        verbose: If True, print progress information.

    Returns:
        Tuple of (mean_parameters, std_parameters).
    """
    n = len(plog)
    plist = pc.array(plist)
    dlist = pc.array(dlist)
    plog = pc.array(plog)

    # Generate leave-one-out resample indices
    resample_indices = [list(range(i)) + list(range(i + 1, n)) for i in range(n)]

    # Generate verbose labels
    verbose_labels = [f"removed index: {i}\np = {plist[i]}, d = {dlist[i]}" for i in range(n)]

    return _jackknife_threshold_core(
        plist,
        dlist,
        plog,
        func,
        p0,
        maxfev,
        resample_indices,
        verbose=verbose,
        verbose_labels=verbose_labels if verbose else None,
    )


def jackknife_p(
    plist: Array[f64] | list[float],
    dlist: Array[f64] | list[float],
    plog: Array[f64] | list[float],
    func: Callable[..., float | Array[f64]],
    p0: Array[f64] | list[float],
    maxfev: int = 100000,
    *,
    verbose: bool = True,
) -> tuple[Array[f64], Array[f64]]:
    """Perform jackknife resampling by removing each unique probability value.

    Args:
        plist: List of probability values.
        dlist: List of distance values.
        plog: List of logical error probabilities.
        func: Fitting function to use.
        p0: Initial parameter guess.
        maxfev: Maximum function evaluations.
        verbose: If True, print progress information.

    Returns:
        Tuple of (mean_parameters, std_parameters).
    """
    plist = pc.array(plist)
    dlist = pc.array(dlist)
    plog = pc.array(plog)

    uplist = sorted(set(plist.tolist()))

    # Generate resample indices for each unique p value
    resample_indices = []
    verbose_labels = []

    for p_val in uplist:
        mask = plist != p_val
        indices = pc.where(mask)[0].tolist()
        resample_indices.append(indices)
        verbose_labels.append(f"removed p: {p_val}")

    return _jackknife_threshold_core(
        plist,
        dlist,
        plog,
        func,
        p0,
        maxfev,
        resample_indices,
        verbose=verbose,
        verbose_labels=verbose_labels if verbose else None,
    )


def jackknife_d(
    plist: Array[f64] | list[float],
    dlist: Array[f64] | list[float],
    plog: Array[f64] | list[float],
    func: Callable[..., float | Array[f64]],
    p0: Array[f64] | list[float],
    maxfev: int = 100000,
    *,
    verbose: bool = True,
) -> tuple[Array[f64], Array[f64]]:
    """Perform jackknife resampling by removing each unique distance value.

    Args:
        plist: List of probability values.
        dlist: List of distance values.
        plog: List of logical error probabilities.
        func: Fitting function to use.
        p0: Initial parameter guess.
        maxfev: Maximum function evaluations.
        verbose: If True, print progress information.

    Returns:
        Tuple of (mean_parameters, std_parameters).
    """
    plist = pc.array(plist)
    dlist = pc.array(dlist)
    plog = pc.array(plog)

    udlist = sorted(set(dlist.tolist()))

    # Generate resample indices for each unique d value
    resample_indices = []
    verbose_labels = []

    for d_val in udlist:
        mask = dlist != d_val
        indices = pc.where(mask)[0].tolist()
        resample_indices.append(indices)
        verbose_labels.append(f"removed d: {d_val}")

    return _jackknife_threshold_core(
        plist,
        dlist,
        plog,
        func,
        p0,
        maxfev,
        resample_indices,
        verbose=verbose,
        verbose_labels=verbose_labels if verbose else None,
    )


def get_est(
    value_is: list[float],
    label: str,
    *,
    verbose: bool = True,
) -> tuple[float, float]:
    """Calculate mean and standard deviation estimate.

    Args:
        value_is: List of values to analyze.
        label: Label for verbose output.
        verbose: If True, print the results.

    Returns:
        Tuple of (mean, standard_deviation).
    """
    v_est = sum(value_is) / len(value_is)
    v_est_std = pc.std(value_is)

    if verbose:
        print(f"{label}_est: {v_est} (mean) +- {v_est_std} (std)")

    return v_est, v_est_std


def get_i(
    result: dict[str, tuple[float, float]],
    symbol: str,
    value_list: list[float],
    *,
    verbose: bool = True,
) -> None:
    """Extract and append a value from results dictionary.

    Args:
        result: Dictionary containing parameter estimates.
        symbol: Key to extract from the results.
        value_list: List to append the extracted value to.
        verbose: If True, print the extracted value.
    """
    value_i = result[symbol][0]
    value_list.append(value_i)

    if verbose:
        print(f"{symbol}_i = {value_i}")



"""Single-qubit square root Pauli gate implementations.

This module provides square root of Pauli gate implementations (√X, √Y, √Z)
which are Clifford gates that when applied twice give the corresponding
Pauli operators, useful in quantum error correction protocols.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from pecos.slr.qeclib.qubit.qgate_base import QGate


class SX(QGate):
    """Square root of Pauli X gate.

    Action on Pauli operators:
    X -> X
    Z -> -Y
    Y -> Z
    """


class SY(QGate):
    """Square root of Pauli Y gate.

    This gate is the square root of the Pauli Y operation.
    """


class SZ(QGate):
    """Square root of Pauli Z gate (S gate).

    This gate performs a π/2 rotation around the Z-axis.
    """


S = SZ


class SXdg(QGate):
    """Inverse square root of Pauli X gate.

    This gate is the inverse of the square root of Pauli X operation.
    """


class SYdg(QGate):
    """Inverse square root of Pauli Y gate.

    This gate is the inverse of the square root of Pauli Y operation.
    """


class SZdg(QGate):
    """Inverse square root of Pauli Z gate (S-dagger gate).

    This gate performs a -π/2 rotation around the Z-axis.
    """


Sdg = SZdg



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""AST node definitions for SLR programs.

This module defines the Abstract Syntax Tree (AST) nodes for representing
SLR quantum programs. The AST provides:

- Clean separation of syntax from semantics
- Uniform interface for analysis and transformation
- Visitor pattern support for traversal
- Integration with QAlloc for qubit slot management

All nodes are immutable frozen dataclasses for safety and hashability.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    from pecos.slr.ast.visitor import AstVisitor, T


# =============================================================================
# Source Location
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class SourceLocation:
    """Source location for error reporting.

    Tracks where in the source a node originated for error messages.
    """

    line: int
    column: int
    file: str | None = None

    def __str__(self) -> str:
        if self.file:
            return f"{self.file}:{self.line}:{self.column}"
        return f"{self.line}:{self.column}"


# =============================================================================
# Base Node
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class AstNode(ABC):
    """Base class for all AST nodes.

    All AST nodes are immutable frozen dataclasses that support:
    - Visitor pattern via accept()
    - Child traversal via children()
    - Optional source location tracking
    """

    location: SourceLocation | None = field(default=None, compare=False, repr=False)

    @abstractmethod
    def accept(self, visitor: AstVisitor[T]) -> T:
        """Accept a visitor for traversal."""
        ...

    def children(self) -> Sequence[AstNode]:
        """Return child nodes for traversal. Override in subclasses."""
        return ()


# =============================================================================
# Enums
# =============================================================================


class GateKind(Enum):
    """All supported quantum gate types."""

    # Single-qubit Paulis
    X = auto()
    Y = auto()
    Z = auto()

    # Hadamard
    H = auto()

    # Phase gates
    S = auto()
    Sdg = auto()
    T = auto()
    Tdg = auto()

    # Square root gates
    SX = auto()
    SY = auto()
    SZ = auto()
    SXdg = auto()
    SYdg = auto()
    SZdg = auto()

    # Rotation gates (parameterized)
    RX = auto()
    RY = auto()
    RZ = auto()

    # Two-qubit Clifford gates
    CX = auto()
    CY = auto()
    CZ = auto()
    CH = auto()

    # Two-qubit rotation gates
    SXX = auto()
    SYY = auto()
    SZZ = auto()
    SXXdg = auto()
    SYYdg = auto()
    SZZdg = auto()
    RZZ = auto()

    # Controlled rotation gates (parameterized)
    CRX = auto()
    CRY = auto()
    CRZ = auto()

    # Face rotations
    F = auto()
    Fdg = auto()
    F4 = auto()
    F4dg = auto()

    @property
    def arity(self) -> int:
        """Number of qubit arguments required."""
        two_qubit = {
            GateKind.CX,
            GateKind.CY,
            GateKind.CZ,
            GateKind.CH,
            GateKind.SXX,
            GateKind.SYY,
            GateKind.SZZ,
            GateKind.SXXdg,
            GateKind.SYYdg,
            GateKind.SZZdg,
            GateKind.RZZ,
            GateKind.CRX,
            GateKind.CRY,
            GateKind.CRZ,
        }
        return 2 if self in two_qubit else 1

    @property
    def is_parameterized(self) -> bool:
        """Whether this gate takes angle parameters."""
        return self in {
            GateKind.RX,
            GateKind.RY,
            GateKind.RZ,
            GateKind.RZZ,
            GateKind.CRX,
            GateKind.CRY,
            GateKind.CRZ,
        }


class BinaryOp(Enum):
    """Binary operators for expressions."""

    # Arithmetic
    ADD = auto()
    SUB = auto()
    MUL = auto()
    DIV = auto()

    # Comparison
    EQ = auto()
    NE = auto()
    LT = auto()
    LE = auto()
    GT = auto()
    GE = auto()

    # Logical
    AND = auto()
    OR = auto()
    XOR = auto()

    # Bitwise
    LSHIFT = auto()
    RSHIFT = auto()


class UnaryOp(Enum):
    """Unary operators for expressions."""

    NOT = auto()
    NEG = auto()


# =============================================================================
# References
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class SlotRef(AstNode):
    """Reference to a qubit slot in an allocator.

    This is the AST equivalent of QubitRef - it identifies a specific
    slot in a named allocator.
    """

    allocator: str
    index: int

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_slot_ref(self)

    def __str__(self) -> str:
        return f"{self.allocator}[{self.index}]"


@dataclass(frozen=True, kw_only=True)
class BitRef(AstNode):
    """Reference to a classical bit in a register."""

    register: str
    index: int

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_bit_ref(self)

    def __str__(self) -> str:
        return f"{self.register}[{self.index}]"


# =============================================================================
# Expressions
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class Expression(AstNode, ABC):
    """Base class for all expressions."""


@dataclass(frozen=True, kw_only=True)
class LiteralExpr(Expression):
    """Literal value (int, float, bool)."""

    value: int | float | bool

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_literal(self)


@dataclass(frozen=True, kw_only=True)
class VarExpr(Expression):
    """Variable reference."""

    name: str

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_var(self)


@dataclass(frozen=True, kw_only=True)
class BitExpr(Expression):
    """Bit reference as expression (for conditions)."""

    ref: BitRef

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_bit_expr(self)

    def children(self) -> Sequence[AstNode]:
        return (self.ref,)


@dataclass(frozen=True, kw_only=True)
class BinaryExpr(Expression):
    """Binary operation."""

    op: BinaryOp
    left: Expression
    right: Expression

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_binary(self)

    def children(self) -> Sequence[AstNode]:
        return (self.left, self.right)


@dataclass(frozen=True, kw_only=True)
class UnaryExpr(Expression):
    """Unary operation."""

    op: UnaryOp
    operand: Expression

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_unary(self)

    def children(self) -> Sequence[AstNode]:
        return (self.operand,)


# =============================================================================
# Type Expressions
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class TypeExpr(AstNode, ABC):
    """Base class for type expressions."""


@dataclass(frozen=True, kw_only=True)
class QubitTypeExpr(TypeExpr):
    """Single qubit type."""

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_qubit_type(self)


@dataclass(frozen=True, kw_only=True)
class BitTypeExpr(TypeExpr):
    """Single classical bit type."""

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_bit_type(self)


@dataclass(frozen=True, kw_only=True)
class ArrayTypeExpr(TypeExpr):
    """Array type with element type and size."""

    element: TypeExpr
    size: int

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_array_type(self)

    def children(self) -> Sequence[AstNode]:
        return (self.element,)


@dataclass(frozen=True, kw_only=True)
class AllocatorTypeExpr(TypeExpr):
    """Qubit allocator type with capacity."""

    capacity: int

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_allocator_type(self)


# =============================================================================
# Declarations
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class Declaration(AstNode, ABC):
    """Base class for all declarations."""


@dataclass(frozen=True, kw_only=True)
class AllocatorDecl(Declaration):
    """Qubit allocator declaration.

    Represents a QAlloc in the AST. Can have a parent for hierarchical
    allocation.
    """

    name: str
    capacity: int
    parent: str | None = None  # Name of parent allocator

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_allocator_decl(self)


@dataclass(frozen=True, kw_only=True)
class RegisterDecl(Declaration):
    """Classical register declaration."""

    name: str
    size: int
    is_result: bool = True

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_register_decl(self)


# =============================================================================
# Statements
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class Statement(AstNode, ABC):
    """Base class for all statements."""


@dataclass(frozen=True, kw_only=True)
class GateOp(Statement):
    """Quantum gate application."""

    gate: GateKind
    targets: tuple[SlotRef, ...]
    params: tuple[Expression, ...] = ()

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_gate(self)

    def children(self) -> Sequence[AstNode]:
        return (*self.targets, *self.params)


@dataclass(frozen=True, kw_only=True)
class PrepareOp(Statement):
    """Prepare qubit slots (unprepared -> prepared).

    Represents the prepare() operation from QAlloc.
    If slots is None, all slots in the allocator are prepared.
    """

    allocator: str
    slots: tuple[int, ...] | None = None  # None means prepare_all

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_prepare(self)


@dataclass(frozen=True, kw_only=True)
class MeasureOp(Statement):
    """Measure qubit slots.

    After measurement, slots transition to unprepared state.
    """

    targets: tuple[SlotRef, ...]
    results: tuple[BitRef, ...] = ()

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_measure(self)

    def children(self) -> Sequence[AstNode]:
        return (*self.targets, *self.results)


@dataclass(frozen=True, kw_only=True)
class AssignOp(Statement):
    """Classical assignment."""

    target: BitRef | str  # Variable name or bit reference
    value: Expression

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_assign(self)

    def children(self) -> Sequence[AstNode]:
        nodes: list[AstNode] = []
        if isinstance(self.target, BitRef):
            nodes.append(self.target)
        nodes.append(self.value)
        return nodes


@dataclass(frozen=True, kw_only=True)
class BarrierOp(Statement):
    """Synchronization barrier."""

    allocators: tuple[str, ...] = ()

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_barrier(self)


@dataclass(frozen=True, kw_only=True)
class CommentOp(Statement):
    """Comment in generated code."""

    text: str

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_comment(self)


@dataclass(frozen=True, kw_only=True)
class ReturnOp(Statement):
    """Return statement."""

    values: tuple[Expression | str, ...] = ()  # Can be variable names

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_return(self)

    def children(self) -> Sequence[AstNode]:
        return tuple(v for v in self.values if isinstance(v, AstNode))


@dataclass(frozen=True, kw_only=True)
class PermuteOp(Statement):
    """Permute qubit register assignments.

    Swaps the qubit assignments between two registers or allocators.
    After Permute(a, b), 'a' refers to what was 'b' and vice versa.

    This is a logical permutation that affects register naming, not
    physical qubit movement.
    """

    sources: tuple[str, ...]  # Initial register/allocator names
    targets: tuple[str, ...]  # Final register/allocator names
    add_comment: bool = True  # Whether to add a comment in generated code

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_permute(self)


# =============================================================================
# Control Flow
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class IfStmt(Statement):
    """Conditional execution."""

    condition: Expression
    then_body: tuple[Statement, ...]
    else_body: tuple[Statement, ...] = ()

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_if(self)

    def children(self) -> Sequence[AstNode]:
        return (self.condition, *self.then_body, *self.else_body)


@dataclass(frozen=True, kw_only=True)
class WhileStmt(Statement):
    """While loop."""

    condition: Expression
    body: tuple[Statement, ...]

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_while(self)

    def children(self) -> Sequence[AstNode]:
        return (self.condition, *self.body)


@dataclass(frozen=True, kw_only=True)
class ForStmt(Statement):
    """For loop with iteration variable."""

    variable: str
    start: Expression
    stop: Expression
    step: Expression | None = None
    body: tuple[Statement, ...] = ()

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_for(self)

    def children(self) -> Sequence[AstNode]:
        nodes: list[AstNode] = [self.start, self.stop]
        if self.step:
            nodes.append(self.step)
        nodes.extend(self.body)
        return nodes


@dataclass(frozen=True, kw_only=True)
class RepeatStmt(Statement):
    """Repeat N times."""

    count: int
    body: tuple[Statement, ...]

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_repeat(self)

    def children(self) -> Sequence[AstNode]:
        return self.body


@dataclass(frozen=True, kw_only=True)
class ParallelBlock(Statement):
    """Parallel execution hint."""

    body: tuple[Statement, ...]

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_parallel(self)

    def children(self) -> Sequence[AstNode]:
        return self.body


# =============================================================================
# Program
# =============================================================================


@dataclass(frozen=True, kw_only=True)
class Program(AstNode):
    """Root node representing an SLR program.

    A program consists of:
    - An optional base allocator (required in strict mode)
    - Declarations (allocators, registers)
    - Body statements
    - Optional return type specification
    """

    name: str
    declarations: tuple[Declaration, ...] = ()
    body: tuple[Statement, ...] = ()
    returns: tuple[TypeExpr, ...] = ()
    allocator: AllocatorDecl | None = None  # Base allocator

    def accept(self, visitor: AstVisitor[T]) -> T:
        return visitor.visit_program(self)

    def children(self) -> Sequence[AstNode]:
        nodes: list[AstNode] = []
        if self.allocator:
            nodes.append(self.allocator)
        nodes.extend(self.declarations)
        nodes.extend(self.body)
        nodes.extend(self.returns)
        return nodes

    def get_allocator(self, name: str) -> AllocatorDecl | None:
        """Find an allocator declaration by name."""
        if self.allocator and self.allocator.name == name:
            return self.allocator
        for decl in self.declarations:
            if isinstance(decl, AllocatorDecl) and decl.name == name:
                return decl
        return None

    def get_register(self, name: str) -> RegisterDecl | None:
        """Find a register declaration by name."""
        for decl in self.declarations:
            if isinstance(decl, RegisterDecl) and decl.name == name:
                return decl
        return None



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""AST to Stim circuit code generator.

This module transforms AST nodes into Stim circuit format.
Stim is a high-performance stabilizer circuit simulator.

Example:
    from pecos.slr.ast import slr_to_ast
    from pecos.slr.ast.codegen import AstToStim

    ast = slr_to_ast(slr_program)
    generator = AstToStim()
    circuit = generator.generate(ast)
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from pecos.slr.ast.nodes import (
    AllocatorDecl,
    BarrierOp,
    ForStmt,
    GateKind,
    GateOp,
    IfStmt,
    MeasureOp,
    ParallelBlock,
    PermuteOp,
    PrepareOp,
    RegisterDecl,
    RepeatStmt,
    WhileStmt,
)

if TYPE_CHECKING:
    import stim

    from pecos.slr.ast.nodes import (
        Program,
        Statement,
    )

# Mapping from AST GateKind to Stim gate names
GATE_TO_STIM: dict[GateKind, str] = {
    # Single-qubit Paulis
    GateKind.X: "X",
    GateKind.Y: "Y",
    GateKind.Z: "Z",
    # Hadamard
    GateKind.H: "H",
    # Phase gates
    GateKind.S: "S",
    GateKind.Sdg: "S_DAG",
    GateKind.T: "T",
    GateKind.Tdg: "T_DAG",
    # Square root gates (mapped to S variants)
    GateKind.SZ: "S",
    GateKind.SZdg: "S_DAG",
    GateKind.SX: "SQRT_X",
    GateKind.SXdg: "SQRT_X_DAG",
    GateKind.SY: "SQRT_Y",
    GateKind.SYdg: "SQRT_Y_DAG",
    # Two-qubit gates
    GateKind.CX: "CX",
    GateKind.CY: "CY",
    GateKind.CZ: "CZ",
    # Two-qubit sqrt gates
    GateKind.SXX: "SQRT_XX",
    GateKind.SYY: "SQRT_YY",
    GateKind.SZZ: "SQRT_ZZ",
    GateKind.SXXdg: "SQRT_XX_DAG",
    GateKind.SYYdg: "SQRT_YY_DAG",
    GateKind.SZZdg: "SQRT_ZZ_DAG",
}

# Two-qubit gate kinds for special handling
TWO_QUBIT_GATES = {
    GateKind.CX,
    GateKind.CY,
    GateKind.CZ,
    GateKind.CH,
    GateKind.SXX,
    GateKind.SYY,
    GateKind.SZZ,
    GateKind.SXXdg,
    GateKind.SYYdg,
    GateKind.SZZdg,
    GateKind.RZZ,
}


@dataclass
class StimCodeGenContext:
    """Context for Stim code generation."""

    qubit_map: dict[tuple[str, int], int] = field(default_factory=dict)
    next_qubit_id: int = 0
    measurement_count: int = 0
    allocator_parents: dict[str, str | None] = field(default_factory=dict)
    allocator_offsets: dict[str, int] = field(default_factory=dict)

    def get_root_allocator(self, name: str) -> str:
        """Get the root allocator for a given allocator name."""
        current = name
        while self.allocator_parents.get(current) is not None:
            current = self.allocator_parents[current]
        return current

    def get_absolute_index(self, allocator: str, index: int) -> int:
        """Get the absolute index in the root allocator."""
        offset = self.allocator_offsets.get(allocator, 0)
        return offset + index

    def get_qubit(self, allocator: str, index: int) -> int:
        """Get or allocate a qubit ID for an allocator slot.

        For child allocators, translates to root allocator with computed offset.
        """
        # Translate to root allocator and absolute index
        root = self.get_root_allocator(allocator)
        abs_index = self.get_absolute_index(allocator, index)

        key = (root, abs_index)
        if key not in self.qubit_map:
            self.qubit_map[key] = self.next_qubit_id
            self.next_qubit_id += 1
        return self.qubit_map[key]


class AstToStim:
    """Transforms AST programs into Stim circuits using recursive descent.

    Generates Stim circuit objects suitable for stabilizer simulation.

    Usage:
        generator = AstToStim()
        circuit = generator.generate(ast_program)
    """

    def __init__(self) -> None:
        """Initialize the generator."""
        self.context = StimCodeGenContext()
        self.circuit: stim.Circuit | None = None

    def generate(self, program: Program) -> stim.Circuit:
        """Generate a Stim circuit for a program.

        Args:
            program: The AST Program to generate code for.

        Returns:
            A stim.Circuit object.
        """
        import stim  # noqa: PLC0415

        self.context = StimCodeGenContext()
        self.circuit = stim.Circuit()

        # Process declarations to allocate qubits
        self._process_declarations(program)

        # Process body statements
        for stmt in program.body:
            self._process_statement(stmt)

        return self.circuit

    def _process_declarations(self, program: Program) -> None:
        """Process declarations to allocate qubits."""
        # First pass: collect allocator parent info
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl):
                self.context.allocator_parents[decl.name] = decl.parent

        if program.allocator:
            self.context.allocator_parents[program.allocator.name] = program.allocator.parent

        # Calculate offsets for child allocators
        self._calculate_allocator_offsets(program)

        # Allocate qubits only for root allocators
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl):
                # Only allocate for root allocators (those without parents)
                if decl.parent is None:
                    for i in range(decl.capacity):
                        self.context.get_qubit(decl.name, i)
            elif isinstance(decl, RegisterDecl):
                pass  # Classical registers don't need qubit allocation

        if program.allocator and program.allocator.parent is None:
            for i in range(program.allocator.capacity):
                self.context.get_qubit(program.allocator.name, i)

    def _calculate_allocator_offsets(self, program: Program) -> None:
        """Calculate the offset of each child allocator within its parent."""
        parent_next_offset: dict[str, int] = {}

        # Root allocators have offset 0
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl) and decl.parent is None:
                self.context.allocator_offsets[decl.name] = 0

        if program.allocator and program.allocator.parent is None:
            self.context.allocator_offsets[program.allocator.name] = 0

        # Process child allocators in declaration order
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl) and decl.parent is not None:
                parent = decl.parent
                if parent not in parent_next_offset:
                    parent_next_offset[parent] = 0

                parent_offset = self.context.allocator_offsets.get(parent, 0)
                self.context.allocator_offsets[decl.name] = parent_offset + parent_next_offset[parent]
                parent_next_offset[parent] += decl.capacity

    def _process_statement(self, stmt: Statement) -> None:
        """Process a statement using recursive descent."""
        if isinstance(stmt, GateOp):
            self._process_gate(stmt)
        elif isinstance(stmt, MeasureOp):
            self._process_measure(stmt)
        elif isinstance(stmt, PrepareOp):
            self._process_prepare(stmt)
        elif isinstance(stmt, BarrierOp):
            self._process_barrier()
        elif isinstance(stmt, IfStmt):
            self._process_if(stmt)
        elif isinstance(stmt, WhileStmt):
            self._process_while(stmt)
        elif isinstance(stmt, ForStmt):
            self._process_for(stmt)
        elif isinstance(stmt, RepeatStmt):
            self._process_repeat(stmt)
        elif isinstance(stmt, ParallelBlock):
            self._process_parallel(stmt)
        elif isinstance(stmt, PermuteOp):
            self._process_permute(stmt)
        # Other statement types (Comment, Assign, Return) don't generate Stim output

    def _process_gate(self, node: GateOp) -> None:
        """Process a gate operation."""
        stim_gate = GATE_TO_STIM.get(node.gate)
        if stim_gate is None:
            # Skip unsupported gates
            return

        if node.gate in TWO_QUBIT_GATES:
            self._process_two_qubit_gate(node, stim_gate)
        else:
            self._process_single_qubit_gate(node, stim_gate)

    def _process_single_qubit_gate(self, node: GateOp, stim_gate: str) -> None:
        """Process a single-qubit gate."""
        qubits = [self.context.get_qubit(t.allocator, t.index) for t in node.targets]
        self.circuit.append_operation(stim_gate, qubits)

    def _process_two_qubit_gate(self, node: GateOp, stim_gate: str) -> None:
        """Process a two-qubit gate."""
        if len(node.targets) >= 2:
            q0 = self.context.get_qubit(
                node.targets[0].allocator,
                node.targets[0].index,
            )
            q1 = self.context.get_qubit(
                node.targets[1].allocator,
                node.targets[1].index,
            )
            self.circuit.append_operation(stim_gate, [q0, q1])
        elif len(node.targets) % 2 == 0:
            # Process pairs
            for i in range(0, len(node.targets), 2):
                q0 = self.context.get_qubit(
                    node.targets[i].allocator,
                    node.targets[i].index,
                )
                q1 = self.context.get_qubit(
                    node.targets[i + 1].allocator,
                    node.targets[i + 1].index,
                )
                self.circuit.append_operation(stim_gate, [q0, q1])

    def _process_measure(self, node: MeasureOp) -> None:
        """Process a measurement operation."""
        qubits = [self.context.get_qubit(t.allocator, t.index) for t in node.targets]
        self.circuit.append_operation("M", qubits)
        self.context.measurement_count += len(qubits)

    def _process_prepare(self, node: PrepareOp) -> None:
        """Process a prepare/reset operation."""
        if node.slots is None:
            return

        qubits = [self.context.get_qubit(node.allocator, slot) for slot in node.slots]
        self.circuit.append_operation("R", qubits)

    def _process_barrier(self) -> None:
        """Process a barrier as TICK."""
        self.circuit.append("TICK")

    def _process_if(self, node: IfStmt) -> None:
        """Process an if statement."""
        # Stim doesn't directly support conditionals
        # Process both branches with TICK markers
        self.circuit.append("TICK")

        for stmt in node.then_body:
            self._process_statement(stmt)

        if node.else_body:
            self.circuit.append("TICK")
            for stmt in node.else_body:
                self._process_statement(stmt)

    def _process_while(self, node: WhileStmt) -> None:
        """Process a while loop."""
        # Stim doesn't support runtime loops - process body once
        self.circuit.append("TICK")
        for stmt in node.body:
            self._process_statement(stmt)

    def _process_for(self, node: ForStmt) -> None:
        """Process a for loop."""
        # Try to unroll if bounds are static
        if isinstance(node.start, int) and isinstance(node.stop, int):
            step = node.step if isinstance(node.step, int) else 1
            for _ in range(node.start, node.stop, step):
                for stmt in node.body:
                    self._process_statement(stmt)
        else:
            # Can't unroll - process body once
            for stmt in node.body:
                self._process_statement(stmt)

    def _process_repeat(self, node: RepeatStmt) -> None:
        """Process a repeat loop using Stim's REPEAT block."""
        import stim  # noqa: PLC0415

        if node.count <= 0:
            return

        # Build sub-circuit for repeat body
        original_circuit = self.circuit
        self.circuit = stim.Circuit()

        for stmt in node.body:
            self._process_statement(stmt)

        sub_circuit = self.circuit
        self.circuit = original_circuit

        # Add repeat block if sub-circuit has content
        if len(sub_circuit) > 0:
            self.circuit.append(stim.CircuitRepeatBlock(node.count, sub_circuit))

    def _process_parallel(self, node: ParallelBlock) -> None:
        """Process a parallel block."""
        # In Stim, operations within a block are naturally parallel
        for stmt in node.body:
            self._process_statement(stmt)

    def _process_permute(self, node: PermuteOp) -> None:
        """Process a permutation operation.

        Updates the internal allocator mapping to swap qubit references.
        Stim doesn't have a permute instruction, so this just updates
        how we map allocator names to qubit indices.
        """
        # Swap the allocator mappings
        for src, tgt in zip(node.sources, node.targets, strict=False):
            # Get current offsets
            src_offset = self.context.allocator_offsets.get(src, 0)
            tgt_offset = self.context.allocator_offsets.get(tgt, 0)
            # Swap them
            self.context.allocator_offsets[src] = tgt_offset
            self.context.allocator_offsets[tgt] = src_offset


def ast_to_stim(program: Program) -> stim.Circuit:
    """Convert an AST Program to a Stim circuit.

    Convenience function for simple code generation.

    Args:
        program: The AST Program to convert.

    Returns:
        A stim.Circuit object.
    """
    generator = AstToStim()
    return generator.generate(program)


def ast_to_stim_str(program: Program) -> str:
    """Convert an AST Program to a Stim circuit string.

    Convenience function for getting string output.

    Args:
        program: The AST Program to convert.

    Returns:
        Stim circuit as a string.
    """
    circuit = ast_to_stim(program)
    return str(circuit)



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

"""Surface code patch with runtime configuration.

Provides a flexible, runtime-configurable surface code patch
with geometry stored as data structures.
"""

from dataclasses import dataclass, field
from enum import Enum, auto
from typing import TYPE_CHECKING

from pecos_rslib.num import zeros

from pecos.qec.surface.layouts import (
    compute_rotated_x_stabilizers,
    compute_rotated_z_stabilizers,
    compute_x_stabilizer_supports,
    compute_z_stabilizer_supports,
    get_rotated_logical_x,
    get_rotated_logical_z,
)

if TYPE_CHECKING:
    import pecos


class PatchOrientation(Enum):
    """Orientation of the surface code patch boundaries."""

    X_TOP_BOTTOM = auto()  # X boundaries on top/bottom, Z on left/right
    Z_TOP_BOTTOM = auto()  # Z boundaries on top/bottom, X on left/right


@dataclass(frozen=True)
class Stabilizer:
    """A stabilizer measurement in the surface code."""

    index: int
    stab_type: str  # 'X' or 'Z'
    data_qubits: tuple[int, ...]
    is_boundary: bool
    position: tuple[int, int] = (0, 0)

    @property
    def weight(self) -> int:
        """Number of data qubits in this stabilizer."""
        return len(self.data_qubits)


@dataclass(frozen=True)
class LogicalOperator:
    """A logical operator for the surface code."""

    op_type: str
    data_qubits: tuple[int, ...]


@dataclass
class PatchGeometry:
    """Geometry of a surface code patch.

    Supports both rotated (default) and standard (non-rotated) surface codes.
    The rotated layout is more common and uses fewer qubits for the same distance.
    """

    dx: int
    dz: int
    orientation: PatchOrientation = PatchOrientation.X_TOP_BOTTOM
    rotated: bool = True

    num_data: int = field(init=False)
    num_x_stab: int = field(init=False)
    num_z_stab: int = field(init=False)

    pos_to_id: dict[tuple[int, int], int] = field(default_factory=dict)
    id_to_pos: dict[int, tuple[int, int]] = field(default_factory=dict)

    x_stabilizers: list[Stabilizer] = field(default_factory=list)
    z_stabilizers: list[Stabilizer] = field(default_factory=list)

    logical_x: LogicalOperator | None = None
    logical_z: LogicalOperator | None = None

    def __post_init__(self) -> None:
        """Initialize computed fields and generate geometry."""
        self.num_data = self.dx * self.dz
        self.num_x_stab = (self.dx * self.dz - 1) // 2
        self.num_z_stab = (self.dx * self.dz - 1) // 2

        self._generate_layout()
        self._generate_stabilizers()
        self._generate_logical_operators()

    def _generate_layout(self) -> None:
        for row in range(self.dx):
            for col in range(self.dz):
                idx = row * self.dz + col
                pos = (row, col)
                self.pos_to_id[pos] = idx
                self.id_to_pos[idx] = pos

    def _generate_stabilizers(self) -> None:
        d = min(self.dx, self.dz)

        if self.rotated:
            x_supports = compute_rotated_x_stabilizers(d)
            z_supports = compute_rotated_z_stabilizers(d)
        else:
            x_supports = compute_x_stabilizer_supports(d)
            z_supports = compute_z_stabilizer_supports(d)

        self.x_stabilizers = [
            Stabilizer(
                index=s.index,
                stab_type="X",
                data_qubits=s.data_qubits,
                is_boundary=s.is_boundary,
            )
            for s in x_supports
        ]

        self.z_stabilizers = [
            Stabilizer(
                index=s.index,
                stab_type="Z",
                data_qubits=s.data_qubits,
                is_boundary=s.is_boundary,
            )
            for s in z_supports
        ]

        self.num_x_stab = len(self.x_stabilizers)
        self.num_z_stab = len(self.z_stabilizers)

    def _generate_logical_operators(self) -> None:
        d = min(self.dx, self.dz)

        if self.rotated:
            logical_x_qubits = get_rotated_logical_x(d)
            logical_z_qubits = get_rotated_logical_z(d)
        else:
            logical_x_qubits = tuple(i * self.dz for i in range(self.dx))
            logical_z_qubits = tuple(range(self.dz))

        self.logical_x = LogicalOperator("X", logical_x_qubits)
        self.logical_z = LogicalOperator("Z", logical_z_qubits)

    @property
    def distance(self) -> int:
        """Code distance (minimum of dx and dz)."""
        return min(self.dx, self.dz)

    @property
    def num_ancilla(self) -> int:
        """Number of ancilla qubits."""
        return 2

    @property
    def num_qubits(self) -> int:
        """Total number of qubits (data + ancilla)."""
        return self.num_data + self.num_ancilla


class SurfacePatch:
    """A configurable surface code patch.

    Supports both rotated (default) and standard (non-rotated) layouts.

    Example:
        >>> patch = SurfacePatch.create(distance=5)  # Rotated (default)
        >>> patch = SurfacePatch.create(distance=5, rotated=False)  # Standard
        >>> patch = SurfacePatch.create(dx=3, dz=5)  # Asymmetric
    """

    def __init__(self, geometry: PatchGeometry) -> None:
        """Initialize a surface patch with the given geometry."""
        self.geometry = geometry

    @classmethod
    def create(
        cls,
        distance: int | None = None,
        dx: int | None = None,
        dz: int | None = None,
        orientation: PatchOrientation = PatchOrientation.X_TOP_BOTTOM,
        *,
        rotated: bool = True,
    ) -> "SurfacePatch":
        """Create a surface code patch.

        Args:
            distance: Symmetric code distance (must be odd >= 3).
            dx: X distance for asymmetric codes.
            dz: Z distance for asymmetric codes.
            orientation: Patch boundary orientation.
            rotated: If True (default), use the rotated layout which is more
                common and uses fewer qubits. If False, use the standard
                (non-rotated) layout.
        """
        if distance is not None:
            if distance < 3 or distance % 2 == 0:
                msg = f"Distance must be odd >= 3, got {distance}"
                raise ValueError(msg)
            dx = dx or distance
            dz = dz or distance
        elif dx is not None and dz is not None:
            if dx < 3 or dx % 2 == 0:
                msg = f"dx must be odd >= 3, got {dx}"
                raise ValueError(msg)
            if dz < 3 or dz % 2 == 0:
                msg = f"dz must be odd >= 3, got {dz}"
                raise ValueError(msg)
        else:
            msg = "Must provide either distance or both dx and dz"
            raise ValueError(msg)

        geometry = PatchGeometry(dx=dx, dz=dz, orientation=orientation, rotated=rotated)
        return cls(geometry)

    @property
    def distance(self) -> int:
        """Code distance (minimum of dx and dz)."""
        return self.geometry.distance

    @property
    def dx(self) -> int:
        """X distance of the patch."""
        return self.geometry.dx

    @property
    def dz(self) -> int:
        """Z distance of the patch."""
        return self.geometry.dz

    @property
    def num_data(self) -> int:
        """Number of data qubits."""
        return self.geometry.num_data

    @property
    def num_qubits(self) -> int:
        """Total number of qubits (data + ancilla)."""
        return self.geometry.num_qubits

    @property
    def x_stabilizers(self) -> list[Stabilizer]:
        """X stabilizers of the patch."""
        return self.geometry.x_stabilizers

    @property
    def z_stabilizers(self) -> list[Stabilizer]:
        """Z stabilizers of the patch."""
        return self.geometry.z_stabilizers

    @property
    def rotated(self) -> bool:
        """True if using rotated layout, False for standard layout."""
        return self.geometry.rotated

    def get_parity_matrix(self, stab_type: str) -> "pecos.Array":
        """Get parity check matrix."""
        stabs = self.x_stabilizers if stab_type == "X" else self.z_stabilizers
        num_stab = len(stabs)
        matrix = zeros((num_stab, self.num_data), dtype="int64")

        for stab in stabs:
            for q in stab.data_qubits:
                matrix[stab.index, q] = 1

        return matrix


class SurfacePatchBuilder:
    """Builder for creating SurfacePatch instances.

    By default, creates a rotated surface code (more common). Use `.standard()`
    to create a non-rotated surface code.

    Example:
        >>> patch = SurfacePatchBuilder().with_distance(5).with_orientation(PatchOrientation.Z_TOP_BOTTOM).build()

        >>> # Non-rotated (standard) surface code:
        >>> patch = SurfacePatchBuilder().with_distance(5).standard().build()
    """

    def __init__(self) -> None:
        """Initialize the builder with default settings."""
        self._distance: int | None = None
        self._dx: int | None = None
        self._dz: int | None = None
        self._orientation: PatchOrientation = PatchOrientation.X_TOP_BOTTOM
        self._rotated: bool = True

    def with_distance(self, distance: int) -> "SurfacePatchBuilder":
        """Set symmetric distance."""
        self._distance = distance
        return self

    def with_distances(self, dx: int, dz: int) -> "SurfacePatchBuilder":
        """Set asymmetric distances."""
        self._dx = dx
        self._dz = dz
        return self

    def with_orientation(self, orientation: PatchOrientation) -> "SurfacePatchBuilder":
        """Set patch orientation."""
        self._orientation = orientation
        return self

    def rotated(self) -> "SurfacePatchBuilder":
        """Use rotated surface code layout (default).

        The rotated layout is more common and uses fewer physical qubits
        for the same code distance.
        """
        self._rotated = True
        return self

    def standard(self) -> "SurfacePatchBuilder":
        """Use standard (non-rotated) surface code layout.

        The standard layout uses more physical qubits but may be preferred
        for certain applications or compatibility with existing code.
        """
        self._rotated = False
        return self

    def build(self) -> SurfacePatch:
        """Build the SurfacePatch."""
        return SurfacePatch.create(
            distance=self._distance,
            dx=self._dx,
            dz=self._dz,
            orientation=self._orientation,
            rotated=self._rotated,
        )



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Converter from SLR representation to AST representation.

This module provides utilities to convert SLR programs (Main, Block, etc.)
into the structured AST representation for analysis and transformation.

Example:
    from pecos.slr import Main, QReg
    from pecos.slr.qeclib import qubit as qb
    from pecos.slr.ast import SLRToAST

    prog = Main(
        q := QReg("q", 2),
        qb.Prep(q[0]),
        qb.H(q[0]),
    )

    converter = SLRToAST()
    ast = converter.convert(prog)
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pecos.slr.ast.nodes import (
    AllocatorDecl,
    ArrayTypeExpr,
    AssignOp,
    BarrierOp,
    BinaryExpr,
    BinaryOp,
    BitExpr,
    BitRef,
    BitTypeExpr,
    CommentOp,
    ForStmt,
    GateKind,
    GateOp,
    IfStmt,
    LiteralExpr,
    MeasureOp,
    ParallelBlock,
    PermuteOp,
    PrepareOp,
    Program,
    QubitTypeExpr,
    RegisterDecl,
    RepeatStmt,
    ReturnOp,
    SlotRef,
    UnaryExpr,
    UnaryOp,
    VarExpr,
    WhileStmt,
)

if TYPE_CHECKING:
    from pecos.slr.ast.nodes import (
        Expression,
        Statement,
        TypeExpr,
    )
    from pecos.slr.block import Block
    from pecos.slr.main import Main


# Mapping from SLR gate class names to AST GateKind
GATE_KIND_MAP: dict[str, GateKind] = {
    # Single-qubit Paulis
    "X": GateKind.X,
    "Y": GateKind.Y,
    "Z": GateKind.Z,
    # Hadamard
    "H": GateKind.H,
    # Phase gates
    "S": GateKind.S,
    "Sdg": GateKind.Sdg,
    "T": GateKind.T,
    "Tdg": GateKind.Tdg,
    # Square root gates
    "SX": GateKind.SX,
    "SY": GateKind.SY,
    "SZ": GateKind.SZ,
    "SXdg": GateKind.SXdg,
    "SYdg": GateKind.SYdg,
    "SZdg": GateKind.SZdg,
    # Rotation gates
    "RX": GateKind.RX,
    "RY": GateKind.RY,
    "RZ": GateKind.RZ,
    # Two-qubit gates
    "CX": GateKind.CX,
    "CNOT": GateKind.CX,  # Alias
    "CY": GateKind.CY,
    "CZ": GateKind.CZ,
    "CH": GateKind.CH,
    # Two-qubit rotation gates
    "SXX": GateKind.SXX,
    "SYY": GateKind.SYY,
    "SZZ": GateKind.SZZ,
    "SXXdg": GateKind.SXXdg,
    "SYYdg": GateKind.SYYdg,
    "SZZdg": GateKind.SZZdg,
    "RZZ": GateKind.RZZ,
    # Controlled rotation gates
    "CRX": GateKind.CRX,
    "CRY": GateKind.CRY,
    "CRZ": GateKind.CRZ,
    # Face rotations
    "F": GateKind.F,
    "Fdg": GateKind.Fdg,
    "F4": GateKind.F4,
    "F4dg": GateKind.F4dg,
}

# Mapping from SLR BinOp class names to AST BinaryOp
BINARY_OP_MAP: dict[str, BinaryOp] = {
    "PLUS": BinaryOp.ADD,
    "MINUS": BinaryOp.SUB,
    "MUL": BinaryOp.MUL,
    "DIV": BinaryOp.DIV,
    "EQUIV": BinaryOp.EQ,
    "NEQUIV": BinaryOp.NE,
    "LT": BinaryOp.LT,
    "LE": BinaryOp.LE,
    "GT": BinaryOp.GT,
    "GE": BinaryOp.GE,
    "AND": BinaryOp.AND,
    "OR": BinaryOp.OR,
    "XOR": BinaryOp.XOR,
    "LSHIFT": BinaryOp.LSHIFT,
    "RSHIFT": BinaryOp.RSHIFT,
}

# Mapping from SLR UnaryOp class names to AST UnaryOp
UNARY_OP_MAP: dict[str, UnaryOp] = {
    "NOT": UnaryOp.NOT,
    "NEG": UnaryOp.NEG,
}


class SlrToAst:
    """Converter from SLR representation to AST.

    Converts SLR Main/Block programs into the structured AST format.
    """

    def __init__(self) -> None:
        """Initialize the converter."""
        self._position = 0  # Track position for source locations

    def convert(self, block: Main | Block) -> Program:
        """Convert an SLR Main/Block to an AST Program.

        Args:
            block: The SLR block to convert.

        Returns:
            An AST Program node.
        """
        self._position = 0

        # Get the block name
        name = getattr(block, "block_name", block.__class__.__name__)

        # Convert declarations from vars
        declarations = self._convert_declarations(block)

        # Convert body statements
        body = self._convert_statements(block.ops)

        # Convert return types if present
        returns = self._convert_return_types(block)

        # Check for base allocator (for new-style allocator-based blocks)
        allocator = None
        for var in block.vars:
            if hasattr(var, "capacity"):  # QAlloc detection
                allocator = AllocatorDecl(
                    name=var.name,
                    capacity=var.capacity,
                    parent=var.parent.name if var.parent else None,
                )
                break

        return Program(
            name=name,
            declarations=tuple(declarations),
            body=tuple(body),
            returns=returns,
            allocator=allocator,
        )

    def _convert_declarations(self, block: Block) -> list:
        """Convert block variables to AST declarations.

        This scans both block.vars (for traditional QReg/CReg declarations)
        and block.ops (for QAllocs created with walrus operator).
        """
        declarations = []
        seen_names: set[str] = set()

        # First, scan block.vars for declarations
        for var in block.vars:
            decl = self._convert_var_to_declaration(var)
            if decl is not None:
                declarations.append(decl)
                seen_names.add(decl.name)

        # Also scan block.ops for QAllocs (they end up in ops when using walrus operator)
        for op in block.ops:
            # Only add if QAlloc and not already seen (avoid duplicates)
            if op.__class__.__name__ == "QAlloc" and op.name not in seen_names:
                decl = self._convert_var_to_declaration(op)
                if decl is not None:
                    declarations.append(decl)
                    seen_names.add(decl.name)

        return declarations

    def _convert_return_types(self, block: Block) -> tuple[TypeExpr, ...]:
        """Convert block return type annotations to AST TypeExpr nodes.

        Args:
            block: The SLR block to extract return types from.

        Returns:
            Tuple of AST TypeExpr nodes representing return types.
        """
        # Import here to avoid circular imports
        from pecos.slr.types import ArrayType, ReturnNotSet  # noqa: PLC0415

        # Check if block has return type annotation
        block_returns = getattr(block.__class__, "block_returns", ReturnNotSet)
        if block_returns is ReturnNotSet:
            return ()

        # Handle None (explicitly no return)
        if block_returns is None:
            return ()

        # Convert single return type
        if isinstance(block_returns, ArrayType):
            return (self._convert_array_type(block_returns),)

        # Convert tuple of return types
        if isinstance(block_returns, tuple):
            return tuple(self._convert_array_type(rt) for rt in block_returns)

        return ()

    def _convert_array_type(self, array_type: Any) -> ArrayTypeExpr:
        """Convert an SLR ArrayType to an AST ArrayTypeExpr.

        Args:
            array_type: The SLR ArrayType to convert.

        Returns:
            An AST ArrayTypeExpr node.
        """
        # Determine element type based on elem_type.name
        elem_name = array_type.elem_type.name
        if elem_name == "Qubit":
            element = QubitTypeExpr()
        elif elem_name == "Bit":
            element = BitTypeExpr()
        else:
            # Default to Qubit for unknown types
            element = QubitTypeExpr()

        return ArrayTypeExpr(element=element, size=array_type.size)

    def _convert_var_to_declaration(self, var: Any):
        """Convert a single variable to an AST declaration."""
        var_class = var.__class__.__name__

        if var_class == "QReg":
            # QReg maps to AllocatorDecl
            return AllocatorDecl(name=var.sym, capacity=var.size)

        if var_class == "CReg":
            # CReg maps to RegisterDecl
            return RegisterDecl(
                name=var.sym,
                size=var.size,
                is_result=getattr(var, "result", True),
            )

        if var_class == "QAlloc":
            # QAlloc maps to AllocatorDecl
            return AllocatorDecl(
                name=var.name,
                capacity=var.capacity,
                parent=var.parent.name if var.parent else None,
            )

        # Unknown variable type - skip
        return None

    def _convert_statements(self, ops: list) -> list[Statement]:
        """Convert a list of SLR operations to AST statements."""
        statements = []
        for op in ops:
            stmt = self._convert_statement(op)
            if stmt is not None:
                # Handle flattening of nested blocks
                if isinstance(stmt, tuple) and len(stmt) == 2 and stmt[0] == "__FLATTEN__":
                    # Flatten the nested statements into this list
                    statements.extend(stmt[1])
                else:
                    statements.append(stmt)
        return statements

    def _convert_statement(self, op: Any) -> Statement | None:
        """Convert a single SLR operation to an AST statement."""
        op_class = op.__class__.__name__

        # Gate operations
        if hasattr(op, "is_qgate") and op.is_qgate:
            return self._convert_gate(op)

        # Control flow
        if op_class == "If":
            return self._convert_if(op)

        if op_class == "While":
            return self._convert_while(op)

        if op_class == "For":
            return self._convert_for(op)

        if op_class == "Repeat":
            return self._convert_repeat(op)

        if op_class == "Parallel":
            return self._convert_parallel(op)

        # Misc statements
        if op_class == "Barrier":
            return self._convert_barrier(op)

        if op_class == "Comment":
            return CommentOp(text=op.txt)

        if op_class == "Return":
            return self._convert_return(op)

        if op_class == "Permute":
            return self._convert_permute(op)

        # Assignment operations
        if op_class == "SET":
            return self._convert_assignment(op)

        # Nested blocks (Block subclasses)
        if hasattr(op, "ops"):
            # This is a nested block - flatten its statements into the parent
            # We return a special marker that _convert_statements will handle
            return ("__FLATTEN__", self._convert_statements(op.ops))

        return None

    def _convert_gate(self, gate: Any) -> Statement:
        """Convert an SLR gate to an AST GateOp, PrepareOp, or MeasureOp."""
        gate_name = gate.sym

        # Handle special operations
        if gate_name == "Prep":
            return self._convert_prep(gate)

        if gate_name == "Measure":
            return self._convert_measure(gate)

        # Regular gate
        gate_kind = GATE_KIND_MAP.get(gate_name)
        if gate_kind is None:
            # Unknown gate - use a fallback or raise
            msg = f"Unknown gate type: {gate_name}"
            raise ValueError(msg)

        # Check if qargs contains tuples of qubit pairs (for multi-qubit gates)
        # Pattern: CX((q1, q2), (q3, q4), ...) where each tuple is one gate application
        if gate.qargs and gate_kind.arity > 1 and isinstance(gate.qargs[0], tuple):
            # Each element of qargs is a tuple of qubits for one gate application
            gates = []
            for qubit_tuple in gate.qargs:
                # Expand any registers in the tuple
                expanded_tuple = self._expand_qubit_args(list(qubit_tuple))
                targets = tuple(self._convert_qubit_ref(q) for q in expanded_tuple)
                params: tuple = ()
                if gate.params:
                    params = tuple(self._convert_expression(p) for p in gate.params)
                gates.append(GateOp(gate=gate_kind, targets=targets, params=params))
            return ("__FLATTEN__", gates)

        # Expand full registers into individual qubits
        expanded_qargs = self._expand_qubit_args(gate.qargs)

        # For single-qubit gates applied to full register, generate multiple gates
        if gate_kind.arity == 1 and len(expanded_qargs) > 1:
            # Return a special marker to generate multiple gate operations
            gates = []
            for q in expanded_qargs:
                target = self._convert_qubit_ref(q)
                params: tuple = ()
                if gate.params:
                    params = tuple(self._convert_expression(p) for p in gate.params)
                gates.append(GateOp(gate=gate_kind, targets=(target,), params=params))
            # Return flattening marker
            return ("__FLATTEN__", gates)

        # Convert targets
        targets = tuple(self._convert_qubit_ref(q) for q in expanded_qargs)

        # Convert parameters if present
        params = ()
        if gate.params:
            params = tuple(self._convert_expression(p) for p in gate.params)

        return GateOp(gate=gate_kind, targets=targets, params=params)

    def _convert_prep(self, gate: Any) -> Statement:
        """Convert an SLR Prep gate to an AST PrepareOp or flattened list."""
        if not gate.qargs:
            msg = "Prep gate has no qubit arguments"
            raise ValueError(msg)

        # Expand full registers into individual qubits
        expanded_qargs = self._expand_qubit_args(gate.qargs)

        if not expanded_qargs:
            msg = "Prep gate has no expanded qubit arguments"
            raise ValueError(msg)

        # Get allocator name and slot indices from expanded qargs
        first_qubit = expanded_qargs[0]
        allocator = (
            first_qubit.reg.sym
            if hasattr(first_qubit, "reg") and hasattr(first_qubit.reg, "sym")
            else (str(first_qubit.reg) if hasattr(first_qubit, "reg") else str(first_qubit))
        )

        slots = tuple(q.index for q in expanded_qargs)

        return PrepareOp(allocator=allocator, slots=slots)

    def _convert_measure(self, gate: Any) -> MeasureOp:
        """Convert an SLR Measure gate to an AST MeasureOp."""
        # Expand full registers into individual qubits
        expanded_qargs = self._expand_qubit_args(gate.qargs)
        targets = tuple(self._convert_qubit_ref(q) for q in expanded_qargs)

        # Convert classical output bits if present
        results: tuple = ()
        if gate.cout:
            expanded_cout = self._expand_bit_args(gate.cout)
            results = tuple(self._convert_bit_ref(b) for b in expanded_cout)

        return MeasureOp(targets=targets, results=results)

    def _expand_qubit_args(self, qargs: list) -> list:
        """Expand qubit arguments, converting full registers to individual qubits.

        Filters out non-qubit arguments like strings (e.g., basis state "Z" in Prep).
        """
        expanded = []
        for q in qargs:
            if isinstance(q, str):
                # Skip string arguments (e.g., basis state in Prep)
                continue
            if isinstance(q, list):
                # This is a slice (list of qubits) - recursively expand
                expanded.extend(self._expand_qubit_args(q))
            elif hasattr(q, "size") and hasattr(q, "elems"):
                # This is a full register (QReg) - expand to all qubits
                expanded.extend(q.elems)
            elif hasattr(q, "index") and isinstance(q.index, int):
                # This is an individual qubit with integer index
                expanded.append(q)
            elif hasattr(q, "reg"):
                # This is an individual qubit reference
                expanded.append(q)
            # Skip other non-qubit arguments
        return expanded

    def _expand_bit_args(self, bits: list) -> list:
        """Expand bit arguments, converting full registers to individual bits."""
        expanded = []
        for b in bits:
            if isinstance(b, list):
                # This is a slice (list of bits) - recursively expand
                expanded.extend(self._expand_bit_args(b))
            elif hasattr(b, "size") and hasattr(b, "elems"):
                # This is a full register (CReg) - expand to all bits
                expanded.extend(b.elems)
            else:
                # This is an individual bit
                expanded.append(b)
        return expanded

    def _convert_qubit_ref(self, qubit: Any) -> SlotRef:
        """Convert an SLR Qubit/QubitRef to an AST SlotRef."""
        # Handle both old Qubit (has .reg) and new QubitRef (has .alloc)
        if hasattr(qubit, "reg"):
            allocator = qubit.reg.sym if hasattr(qubit.reg, "sym") else str(qubit.reg)
        elif hasattr(qubit, "alloc"):
            allocator = qubit.alloc.name
        else:
            allocator = str(qubit)

        return SlotRef(allocator=allocator, index=qubit.index)

    def _convert_bit_ref(self, bit: Any) -> BitRef:
        """Convert an SLR Bit to an AST BitRef."""
        register = bit.reg.sym if hasattr(bit.reg, "sym") else str(bit.reg)
        return BitRef(register=register, index=bit.index)

    def _convert_if(self, op: Any) -> IfStmt:
        """Convert an SLR If block to an AST IfStmt."""
        condition = self._convert_expression(op.cond)
        then_body = tuple(self._convert_statements(op.ops))

        else_body: tuple = ()
        if op.else_block is not None:
            else_body = tuple(self._convert_statements(op.else_block.ops))

        return IfStmt(condition=condition, then_body=then_body, else_body=else_body)

    def _convert_while(self, op: Any) -> WhileStmt:
        """Convert an SLR While block to an AST WhileStmt."""
        condition = self._convert_expression(op.cond)
        body = tuple(self._convert_statements(op.ops))

        return WhileStmt(condition=condition, body=body)

    def _convert_for(self, op: Any) -> ForStmt:
        """Convert an SLR For block to an AST ForStmt."""
        variable = str(op.var) if not isinstance(op.var, str) else op.var

        # Handle range-based or explicit start/stop
        if op.iterable is not None:
            # Range object
            if isinstance(op.iterable, range):
                start = LiteralExpr(value=op.iterable.start)
                stop = LiteralExpr(value=op.iterable.stop)
                step_val = op.iterable.step
                step = LiteralExpr(value=step_val) if step_val != 1 else None
            else:
                # Some other iterable - convert bounds
                start = LiteralExpr(value=0)
                stop = LiteralExpr(value=len(op.iterable))
                step = None
        else:
            start = self._convert_expression(op.start)
            stop = self._convert_expression(op.stop)
            step = self._convert_expression(op.step) if op.step is not None and op.step != 1 else None

        body = tuple(self._convert_statements(op.ops))

        return ForStmt(variable=variable, start=start, stop=stop, step=step, body=body)

    def _convert_repeat(self, op: Any) -> RepeatStmt:
        """Convert an SLR Repeat block to an AST RepeatStmt."""
        count = op.cond  # Repeat uses cond to store the count
        body = tuple(self._convert_statements(op.ops))

        return RepeatStmt(count=count, body=body)

    def _convert_parallel(self, op: Any) -> ParallelBlock:
        """Convert an SLR Parallel block to an AST ParallelBlock."""
        body = tuple(self._convert_statements(op.ops))
        return ParallelBlock(body=body)

    def _convert_barrier(self, op: Any) -> BarrierOp:
        """Convert an SLR Barrier to an AST BarrierOp."""
        allocators: tuple = ()
        if op.qregs:
            allocators = tuple(
                q.sym if hasattr(q, "sym") else str(q)
                for q in op.qregs
                if not hasattr(q, "index")  # Skip individual qubits
            )
        return BarrierOp(allocators=allocators)

    def _convert_return(self, op: Any) -> ReturnOp:
        """Convert an SLR Return to an AST ReturnOp."""
        values: list = []
        for var in op.return_vars:
            if isinstance(var, str):
                values.append(var)
            elif hasattr(var, "sym"):
                values.append(var.sym)
            else:
                # Try to convert as expression
                values.append(self._convert_expression(var))

        return ReturnOp(values=tuple(values))

    def _convert_permute(self, op: Any) -> PermuteOp:
        """Convert an SLR Permute to an AST PermuteOp."""
        elems_i = op.elems_i
        elems_f = op.elems_f
        add_comment = getattr(op, "comment", True)

        # Check for whole register swap: Permute(reg_a, reg_b)
        # This is a special case where both arguments are single register objects
        if (
            hasattr(elems_i, "sym")
            and hasattr(elems_f, "sym")
            and not hasattr(elems_i, "__iter__")
            and not hasattr(elems_f, "__iter__")
        ):
            # Whole register swap: a <-> b
            # Represented as sources=[a, b], targets=[b, a]
            return PermuteOp(
                sources=(elems_i.sym, elems_f.sym),
                targets=(elems_f.sym, elems_i.sym),
                add_comment=add_comment,
            )

        # Extract register/allocator names from sources (elems_i)
        sources: list[str] = []
        if hasattr(elems_i, "__iter__") and not isinstance(elems_i, str):
            for elem in elems_i:
                if hasattr(elem, "sym"):
                    sources.append(elem.sym)
                elif hasattr(elem, "name"):
                    sources.append(elem.name)
                else:
                    sources.append(str(elem))
        elif hasattr(elems_i, "sym"):
            sources.append(elems_i.sym)
        elif hasattr(elems_i, "name"):
            sources.append(elems_i.name)
        else:
            sources.append(str(elems_i))

        # Extract register/allocator names from targets (elems_f)
        targets: list[str] = []
        if hasattr(elems_f, "__iter__") and not isinstance(elems_f, str):
            for elem in elems_f:
                if hasattr(elem, "sym"):
                    targets.append(elem.sym)
                elif hasattr(elem, "name"):
                    targets.append(elem.name)
                else:
                    targets.append(str(elem))
        elif hasattr(elems_f, "sym"):
            targets.append(elems_f.sym)
        elif hasattr(elems_f, "name"):
            targets.append(elems_f.name)
        else:
            targets.append(str(elems_f))

        return PermuteOp(
            sources=tuple(sources),
            targets=tuple(targets),
            add_comment=add_comment,
        )

    def _convert_assignment(self, op: Any) -> AssignOp:
        """Convert an SLR SET operation to an AST AssignOp."""
        # Target
        target = op.left
        if hasattr(target, "reg") and hasattr(target, "index"):
            # It's a Bit reference
            target = self._convert_bit_ref(target)
        elif hasattr(target, "sym"):
            target = target.sym
        else:
            target = str(target)

        # Value
        value = self._convert_expression(op.right)

        return AssignOp(target=target, value=value)

    def _convert_expression(self, expr: Any) -> Expression:
        """Convert an SLR expression to an AST Expression."""
        if expr is None:
            return LiteralExpr(value=0)

        # Literal values
        if isinstance(expr, bool | int | float):
            return LiteralExpr(value=expr)

        expr_class = expr.__class__.__name__

        # Binary operations
        if expr_class in BINARY_OP_MAP:
            return BinaryExpr(
                op=BINARY_OP_MAP[expr_class],
                left=self._convert_expression(expr.left),
                right=self._convert_expression(expr.right),
            )

        # Unary operations
        if expr_class in UNARY_OP_MAP:
            return UnaryExpr(
                op=UNARY_OP_MAP[expr_class],
                operand=self._convert_expression(expr.value),
            )

        # Bit reference as expression
        if hasattr(expr, "reg") and hasattr(expr, "index") and expr.__class__.__name__ == "Bit":
            return BitExpr(ref=self._convert_bit_ref(expr))

        # Variable reference
        if hasattr(expr, "sym"):
            return VarExpr(name=expr.sym)

        # String variable name
        if isinstance(expr, str):
            return VarExpr(name=expr)

        # Tuple handling for parameter access
        if isinstance(expr, tuple):
            if len(expr) == 1:
                return self._convert_expression(expr[0])
            # Multiple values - return first for now
            return self._convert_expression(expr[0])

        # Fallback - try to get a value
        if hasattr(expr, "value"):
            return LiteralExpr(value=expr.value)

        # Unknown expression type - convert to string var
        return VarExpr(name=str(expr))


def slr_to_ast(block: Main | Block) -> Program:
    """Convert an SLR Main/Block to an AST Program.

    Convenience function for simple conversions.

    Args:
        block: The SLR block to convert.

    Returns:
        An AST Program node.
    """
    converter = SlrToAst()
    return converter.convert(block)
